/************************************************************************/
/* pfm_dns.c - DNS packet filter module					*/
/************************************************************************/
/*
 *  (c)2021,2022 Todd Shadburn <tshadburn@opencomputingsolutions.com>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
/************************************************************************/


//#define NDEBUG
#include <assert.h>

#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <string.h>
#include <errno.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <time.h>
#include <cutil/cutil.h>
#include <pcre.h>

#include <pnetbuffer.h>
#include <logging.h>
#include <udp.h>
#include <tcp.h>
#include <pfm.h>
#include <metric.h>
#include <helpers.h>

#include <pfm_dns.h>
#include <filterset.h>


#define PACKETBUF_INITIAL_ALLOC_COUNT	16


const char *vote_string[] = {
	"pass",
	"deny",
	"drop",
	"log",
};


policy_item_t policy_table[] = {
	{"deny-multiquery",	POLICY_DENY_MULTIPLE_QUERIES},
	{NULL,			POLICY_NOOP},
};


#define PFM_DNS_STAGE_COUNT	3
const char *stage_string[] = {
	"source",
	"query",
	"answer",
};


// Forward prototypes
int vector_process_command(pfm_t *pfm,
	pnetbuffer_t *indata,
	pnetbuffer_t *outdata);
int vector_process_data(pfm_t *pfm,
	packetbuf_t *in_pb);
int udp_backend_io_callback(int fd,
	int events,
	connection_data_t *cd,
	void *arg);
int tcp_server_io_callback(int fd,
	int events,
	connection_data_t *cd,
	void *arg);
int udp_frontend_io_callback(int fd,
	int events,
	connection_data_t *cd,
	void *arg);
int tcp_frontend_io_callback(int fd,
	int events,
	connection_data_t *cd,
	void *arg);
int backend_io_callback(int fd,
	int events,
	connection_data_t *cd,
	void *arg);

packetbuf_t *packetbuf_get_or_alloc(void);
void packetbuf_free(packetbuf_t *pb);


// Module Globals
pfm_vectors_t vectors = {
	.process_command	= vector_process_command,
	.process_data		=  NULL,
};

uint16_t port_list[] = {
	PFM_DNS_PORT_DEFAULT,
	0,
};

pfm_t pfm_info = {
	.version	= PFM_VERSION_1,
	.flags		= 0,
	.name		= "dns",
	.vectors	= &vectors,
	.userdata	= NULL,
};


pfm_dns_t	priv_data;



int backend_add(backend_protocol protocol,
	char *hostname,
	char *service,
	int minconns,
	int maxconns)
{
	int			rc;
	backend_t		*be;

	if (priv_data.backend_count[protocol] == DNS_BACKENDS_MAX)
		return E2BIG;

	pfm_vector_log(priv_data.module,
		LOG_LEVEL_NOTICE,
		"New backend requested for %s:%s:%s with connection counts"
		" min=%d max=%d",
		backend_protocol_to_name(protocol),
		hostname,
		service,
		minconns,
		maxconns);

	be = backend_new(protocol,
			hostname,
			service,
			minconns,
			maxconns);
	if (!be) {
		return ENOMEM;
	}
	priv_data.backend[protocol][priv_data.backend_count[protocol]] = be;
	priv_data.backend_count[protocol]++;

	rc = backend_increase_connections(be, minconns);
	if (rc != 0)
		return rc;
	return 0;
}


int backend_remove(backend_protocol protocol,
	char *hostname,
	char *service)
{
	int			i;
	int			be_oset;
	int			be_last;
	backend_t		*be = NULL;

	pfm_vector_log(priv_data.module,
		LOG_LEVEL_NOTICE,
		"Remove backend requested for %s:%s",
		hostname,
		service);

	// Find old backend in table
	for (i=0 ; i<priv_data.backend_count[protocol] ; i++) {
		be = priv_data.backend[protocol][i];
		if (!strcmp(be->dest_host, hostname)
			&& !strcmp(be->dest_service, service) 
			&& be->protocol == protocol)
			break;
	}
	if (i == priv_data.backend_count[protocol])
		return ENOENT;

	// get references
	be_oset = i;

	// remove backend from table (slide down)
	be_last = priv_data.backend_count[protocol] - 1;
	for (i=be_oset ; i<be_last ; i++) {
		memcpy(&priv_data.backend[protocol][i],
			&priv_data.backend[protocol][i+1],
			sizeof(backend_t*));
	}
	priv_data.backend[protocol][be_last] = NULL;
	priv_data.backend_count[protocol]--;

	// free backend resources
	(void)backend_free(be);

	return 0;
}



int pfm_module_info(pfm_t *pfm)
{
	pfm->version = pfm_info.version;
	pfm->flags = pfm_info.flags;
	pfm->name = pfm_info.name;
	pfm->vectors = pfm_info.vectors;
	pfm->userdata = pfm_info.userdata;
	return 0;
}


int pfm_module_open(pfm_module_t *module)
{
	int			rc;
	int			i;
	int			fd;
	ioctl_opt_register_fd_t	opt;
	ioctl_opt_get_ci_t	o_gci;
	char			cfgstr[RESPONSE_STRING_SIZE];
	packetbuf_t		*pb;
	fd_t			*fdte;

	memset(&priv_data, 0, sizeof(pfm_dns_t));
	priv_data.module = (pfm_module_t *)module;

	// Allocate FD tracking table
	pfm_vector_log(priv_data.module,
		LOG_LEVEL_DEBUG,
		"%s() memory allocation line=%d",
		__func__,
		__LINE__);
	priv_data.fd_table_size = DNS_FD_TABLE_SIZE_DEFAULT;
	priv_data.fd_table = calloc(priv_data.fd_table_size,
					sizeof(fd_t));
	if (!priv_data.fd_table) {
		rc = errno;
		goto pfm_module_open_exit;
	}

	// assign default port(s)
	pfm_info.port_binds = port_list;

	// allocate command buffer array
	pfm_vector_log(priv_data.module,
		LOG_LEVEL_DEBUG,
		"%s() memory allocation line=%d",
		__func__,
		__LINE__);
	priv_data.command_array = calloc(PFM_COMMAND_TOKEN_MAX,
					sizeof(char*));
	if (!priv_data.command_array) {
		rc = errno;
		goto pfm_module_open_exit;
	}
	for (i=0 ; i<PFM_COMMAND_TOKEN_MAX ; i++) {
		priv_data.command_array[i] = calloc(1,
					PFM_COMMAND_STR_SIZE);
		if (!priv_data.command_array[i]) {
			rc = errno;
			goto pfm_module_open_exit;
		}
	}

	// allocate response buffer array
	pfm_vector_log(priv_data.module,
		LOG_LEVEL_DEBUG,
		"%s() memory allocation line=%d",
		__func__,
		__LINE__);
	priv_data.response_array = calloc(RESPONSE_ARRAY_SIZE,
					sizeof(char*));
	if (!priv_data.response_array) {
		rc = errno;
		goto pfm_module_open_exit;
	}
	for (i=0 ; i<RESPONSE_ARRAY_SIZE ; i++) {
		priv_data.response_array[i] = calloc(1,
					RESPONSE_STRING_SIZE);
		if (!priv_data.response_array[i]) {
			rc = errno;
			goto pfm_module_open_exit;
		}
	}
	pfm_vector_log(priv_data.module,
		LOG_LEVEL_DEBUG,
		"Successfully allocated module structures");

	// allocate initial packetbuf free list
	dlist_init(&priv_data.packetbuf_free_head);
	for (i=0 ; i<PACKETBUF_INITIAL_ALLOC_COUNT ; i++) {
		pb = calloc(1, sizeof(packetbuf_t));
		if (!pb) {
			rc = ENOMEM;
			goto pfm_module_open_exit;
		}
		pb->nb = pnetbuffer_init(pb->nb_buffer, PACKET_BUFFER_SIZE);
		dlist_add(&priv_data.packetbuf_free_head,
			(dlist_t*)pb);
	}

	// setup metric table
	metric_init(&priv_data.metric[METRIC_QUERY_TOTAL],
		"dns.query.total",
		METRIC_TYPE_UINT32);
	metric_init(&priv_data.metric[METRIC_QUERY_FILTERED],
		"dns.query.filtered",
		METRIC_TYPE_UINT32);
	metric_init(&priv_data.metric[METRIC_ANSWER_TOTAL],
		"dns.answer.total",
		METRIC_TYPE_UINT32);
	metric_init(&priv_data.metric[METRIC_ANSWER_FILTERED],
		"dns.answer.filtered",
		METRIC_TYPE_UINT32);
	metric_init(&priv_data.metric[METRIC_TCP_CONN_TOTAL],
		"dns.tcp.connections.total",
		METRIC_TYPE_UINT32);
	metric_init(&priv_data.metric[METRIC_TCP_CONN_SERVER_DROP],
		"dns.tcp.connections.server.drop",
		METRIC_TYPE_UINT32);
	metric_init(&priv_data.metric[METRIC_TCP_BACKEND_RECONNECT],
		"dns.tcp.backend.reconnect",
		METRIC_TYPE_UINT32);
	metric_init(&priv_data.metric[METRIC_UDP_CONN_SERVER_DROP],
		"dns.udp.connections.server.drop",
		METRIC_TYPE_UINT32);
	metric_init(&priv_data.metric[METRIC_CONN_CURRENT],
		"dns.connections.concurrent",
		METRIC_TYPE_UINT32);

	// get service bind address
	o_gci.ci_id = PFM_CI_BIND_ADDRESS;
	o_gci.value.str = NULL;
	rc = pfm_vector_ioctl(priv_data.module,
		PFM_IOCTL_GET_CONFIG_ITEM,
		&o_gci);
	if (rc || !o_gci.value.str) {
		pfm_vector_log(priv_data.module,
			LOG_LEVEL_ERROR,
			"Failed to retrieve service bind address"
			" rc=%d",
			rc);
		return rc;
	}
	priv_data.service_bind_address = strdup(o_gci.value.str);
	pfm_vector_log(priv_data.module,
		LOG_LEVEL_INFO,
		"Retrieved service bind address is %s",
		priv_data.service_bind_address);

	// create listener sockets and connect to ioloop
	i = 0;
	while (1) {
		if (port_list[i] == 0) break;
		rc = snprintf(cfgstr,
			RESPONSE_STRING_SIZE,
			"%d",
			port_list[i]);

		// UDP port
		fd = udp_bind(priv_data.service_bind_address, cfgstr);
		if (fd == -1) {
			rc = errno;
			pfm_vector_log(priv_data.module,
				LOG_LEVEL_ERROR,
				"Failed to create server socket for"
				" port=%d/udp rc=%d",
				port_list[i],
				rc);
			return rc;
		}
		fdte = &priv_data.fd_table[fd];
		fd_entry_init(fdte,
			DNS_CONN_TYPE_UDP,
			DNS_USAGE_TYPE_FRONTEND,
			NULL);

		// setup connection structure
		pfm_vector_log(priv_data.module,
			LOG_LEVEL_DEBUG,
			"%s() memory allocation line=%d",
			__func__,
			__LINE__);

		opt.fd = fd;
		opt.port = port_list[i];
		opt.callback = udp_frontend_io_callback;
		opt.callback_arg = NULL;
		rc = pfm_vector_ioctl(priv_data.module,
			PFM_IOCTL_REGISTER_BACKEND_FD,
			&opt);
		if (rc) {
			pfm_vector_log(priv_data.module,
				LOG_LEVEL_ERROR,
				"Failed to register server FD with ioloop"
				" subsystem rc=%d",
				rc);
			i = udp_shutdown(fd);
			return rc;
		}
		pfm_vector_log(priv_data.module,
			LOG_LEVEL_NOTICE,
			"Bound new server socket fd=%d port=%d/udp",
			fd,
			port_list[i]);

		// TCP port
		fd = tcp_bind(priv_data.service_bind_address, cfgstr);
		if (fd == -1) {
			rc = errno;
			pfm_vector_log(priv_data.module,
				LOG_LEVEL_ERROR,
				"Failed to create server socket for"
				" port=%d/tcp rc=%d",
				port_list[i],
				rc);
			return rc;
		}
		fdte = &priv_data.fd_table[fd];
		fd_entry_init(fdte,
			DNS_CONN_TYPE_TCP,
			DNS_USAGE_TYPE_SERVER,
			NULL);

		// setup connection structure
		pfm_vector_log(priv_data.module,
			LOG_LEVEL_DEBUG,
			"%s() memory allocation line=%d",
			__func__,
			__LINE__);

		opt.fd = fd;
		opt.port = port_list[i];
		opt.callback = tcp_server_io_callback;
		opt.callback_arg = NULL;
		rc = pfm_vector_ioctl(priv_data.module,
			PFM_IOCTL_REGISTER_BACKEND_FD,
			&opt);
		if (rc) {
			pfm_vector_log(priv_data.module,
				LOG_LEVEL_ERROR,
				"Failed to register server FD with ioloop"
				" subsystem rc=%d",
				rc);
			i = tcp_shutdown(fd);
			return rc;
		}
		pfm_vector_log(priv_data.module,
			LOG_LEVEL_NOTICE,
			"Bound new server socket fd=%d port=%d/tcp",
			fd,
			port_list[i]);

		i++;
	}

	return 0;

pfm_module_open_exit:
	if (priv_data.command_array)
		free(priv_data.command_array);
	if (priv_data.response_array)
		free(priv_data.response_array);
	pfm_vector_log(priv_data.module,
		LOG_LEVEL_ERROR,
		"%s() Failed an initialization step rc=%d",
		__func__,
		rc);
	return rc;
}


int pfm_module_close(void *args)
{
	int		rc;
	int		i;
	int		p;

	for (p = 0 ; p<_DNS_CONN_TYPE_MAX ; p++) {
		for (i=0 ; i<priv_data.backend_count[p] ; i++) {
		rc = backend_remove(p,
			priv_data.backend[p][i]->dest_host,
			priv_data.backend[p][i]->dest_service);
		pfm_vector_log(priv_data.module,
			LOG_LEVEL_DEBUG,
			"%s() backend removal rc=%d",
			__func__,
			rc);
		}
	}

	if (priv_data.command_array)
		free(priv_data.command_array);
	if (priv_data.fd_table)
		free(priv_data.fd_table);
	return 0;
}



policy_item_t *policy_item_lookup_by_name(char *name)
{
	int		i = 0;
	while (policy_table[i].name != NULL) {
		if (!strcmp(name, policy_table[i].name))
			return &policy_table[i];
		i++;
	}
	return NULL;
}


filterset_t *fslist_get_by_name(SList *list, char *name)
{
	SList		*item;
	filterset_t	*fs;
	item = list;
	while (item) {
		fs = slist_get_data(item);
		if (!strcmp(fs->name, name))
			return fs;
		item = slist_get_next(item);
	}
	return NULL;
}



int list_filterset_cb(filterset_t *fs, void *arg)
{
	pnetbuffer_t	*outdata;
	char		**strarray;

	outdata = arg;
	strarray = calloc(2, sizeof(char*));
	strarray[0] = calloc(256, sizeof(char));
	snprintf(strarray[0],
		256,
		"%s %s %s\n",
		fs->name,
		filterset_type_to_string(fs->type),
		filterset_action_to_string(fs->action));
	(void)pnetbuffer_add_strings(outdata,
			strarray,
			1,
			PNETBUFFER_OPT_STRINGS_COUNT_FIRST);
	free(strarray[0]);
	free(strarray);
	return 0;
}

int list_filter_entry_cb(filter_t *filter, void *arg)
{
	pnetbuffer_t	*outdata;
	char		**strarray;

	outdata = arg;
	strarray = calloc(2, sizeof(char*));
	strarray[0] = calloc(256, sizeof(char));
	snprintf(strarray[0],
		256,
		"%d %s %s \"%s\"\n",
		metric_get_u32(&filter->hits),
		filterset_type_to_string(filter->type),
		filterset_action_to_string(filter->action),
		filter->match);
	(void)pnetbuffer_add_strings(outdata,
			strarray,
			1,
			PNETBUFFER_OPT_STRINGS_COUNT_FIRST);
	free(strarray[0]);
	free(strarray);
	return 0;
}


static inline void dump_packetbuf(packetbuf_t *pb)
{
	pfm_vector_log(priv_data.module,
		LOG_LEVEL_DEBUG,
		"%s() pb=%p",
		__func__,
		pb);
	if (!pb) return;
	pfm_vector_log(priv_data.module,
		LOG_LEVEL_DEBUG,
		"%s() .list=%p .sfd=%d .req=%p",
		__func__,
		pb->list,
		pb->sfd,
		pb->req
	);
	return;
}


static inline void dump_fdte(fd_t *fdte)
{
	struct sockaddr_in      *sin;

	pfm_vector_log(priv_data.module,
		LOG_LEVEL_DEBUG,
		"%s() fdte=%p",
		__func__,
		fdte);
	if (!fdte) return;
	if (fdte->conntype == DNS_CONN_TYPE_TCP) {
		sin = (struct sockaddr_in *)&fdte->saddr;
		pfm_vector_log(priv_data.module,
			LOG_LEVEL_DEBUG,
			"%s() .conntype=%d .usagetype=%d .backend=%p .in_pb=%p"
			" .in_rem=%d .out_pb=%p out_qhead.=%p saddr=%s",
			__func__,
			fdte->conntype,
			fdte->usagetype,
			fdte->backend,
			fdte->in_pb,
			fdte->in_rem,
			fdte->out_pb,
			fdte->out_qhead,
			inet_ntoa(sin->sin_addr)
		);
	} else {
		pfm_vector_log(priv_data.module,
			LOG_LEVEL_DEBUG,
			"%s() .conntype=%d .usagetype=%d .backend=%p .in_pb=%p"
			" .in_rem=%d .out_pb=%p out_qhead.=%p",
			__func__,
			fdte->conntype,
			fdte->usagetype,
			fdte->backend,
			fdte->in_pb,
			fdte->in_rem,
			fdte->out_pb,
			fdte->out_qhead
		);
	}
	dump_packetbuf(fdte->in_pb);
	dump_packetbuf(fdte->out_pb);
	return;
}



int process_subcommand_filterset(pfm_t *pfm,
				pnetbuffer_t *outdata,
				char **token,
				int tokenlen)
{
	int		rc;
	int		toki;
	filter_type	type;
	filter_action	action;
	char		*fname;
	char		*ftype;
	char		*faction;
	char		*fmatch;
	filterset_t	*fs;
	SList		*item;

	toki = 0;
	if (!strcmp(token[toki], "create")) {
		if (tokenlen != 7) {
			rc = PFM_ERR_CMD_INVALID;
			goto process_subcommand_filterset_err;
		}
		// filterset name
		toki++;
		if (strcmp(token[toki], "name")) {
			rc = PFM_ERR_CMD_INVALID;
			goto process_subcommand_filterset_err;
		}
		toki++;
		fname = token[toki];
		fs = fslist_get_by_name(priv_data.filterset_list,
					fname);
		if (fs) {
			rc = PFM_ERR_CMD_INVALID; // TODO PFM_ERR_EEXISTS
			goto process_subcommand_filterset_err;
		}
		// filterset type
		toki++;
		if (strcmp(token[toki], "type")) {
			rc = PFM_ERR_CMD_INVALID;
			goto process_subcommand_filterset_err;
		}
		toki++;
		ftype = token[toki];
		type = filterset_string_to_type(ftype);
		if (type == F_TYPE_NONE) {
			rc = PFM_ERR_CMD_INVALID;
			goto process_subcommand_filterset_err;
		}
		// filterset action
		toki++;
		if (strcmp(token[toki], "action")) {
			rc = PFM_ERR_CMD_INVALID;
			goto process_subcommand_filterset_err;
		}
		toki++;
		faction = token[toki];
		action = filterset_string_to_action(faction);
		if (action == F_ACTION_NONE) {
			rc = PFM_ERR_CMD_INVALID;
			goto process_subcommand_filterset_err;
		}

		fs = filterset_new(fname,
				type,
				action);
		if (!fs) {
			rc = PFM_ERR_MOD_INTERNAL;
			goto process_subcommand_filterset_err;
		}
		rc = slist_item_add(&priv_data.filterset_list, fs);
		if (rc) {
			rc = filterset_destroy(fs);
			rc = PFM_ERR_MOD_INTERNAL;
			goto process_subcommand_filterset_err;
		}
		(void)pfm_pnetbuffer_add_rc_string(outdata,
						PFM_ERR_SUCCESS);
		pfm_vector_log(priv_data.module,
			LOG_LEVEL_NOTICE,
			"Filterset %s created successfully",
			fname);

	} else if (!strcmp(token[toki], "entry")) {
		if (tokenlen != 6) {
			rc = PFM_ERR_CMD_INVALID;
			goto process_subcommand_filterset_err;
		}
		// entry action
		toki++;
		if (!strcmp(token[toki], "add") &&
		    !strcmp(token[toki], "remove")) {
			rc = PFM_ERR_CMD_INVALID;
			goto process_subcommand_filterset_err;
		}
		faction = token[toki];
		// filterset name
		toki++;
		if (strcmp(token[toki], "name")) {
			rc = PFM_ERR_CMD_INVALID;
			goto process_subcommand_filterset_err;
		}
		toki++;
		fname = token[toki];
		fs = fslist_get_by_name(priv_data.filterset_list,
					fname);
		if (!fs) {
			rc = PFM_ERR_NO_ENTRY;
			goto process_subcommand_filterset_err;
		}
		// filter match
		toki++;
		if (strcmp(token[toki], "match")) {
			rc = PFM_ERR_CMD_INVALID;
			goto process_subcommand_filterset_err;
		}
		toki++;
		fmatch = token[toki];
		rc = 0;
		if (!strcmp(faction, "add")) {
			rc = filterset_add_entry(fs,
						fmatch,
						1);
		} else if (!strcmp(faction, "remove")) {
			rc = filterset_remove_entry(fs,
						fmatch);
		}
		if (rc) {
			rc = PFM_ERR_MOD_INTERNAL;
			goto process_subcommand_filterset_err;
		}
		(void)pfm_pnetbuffer_add_rc_string(outdata,
						PFM_ERR_SUCCESS);
		pfm_vector_log(priv_data.module,
			LOG_LEVEL_NOTICE,
			"Filterset %s %s filter command completed"
			" successfully",
			fname,
			faction);


//	} else if (!strcmp(token[toki], "remove")) {

	} else if (!strcmp(token[toki], "list")) {
		(void)pfm_pnetbuffer_add_rc_string(outdata,
						PFM_ERR_SUCCESS);
		item = priv_data.filterset_list;
		while (item) {
			fs = slist_get_data(item);
			list_filterset_cb(fs, outdata);
			item = slist_get_next(item);
		}

	} else if (!strcmp(token[toki], "show")) {
		if (tokenlen != 3) {
			rc = PFM_ERR_CMD_INVALID;
			goto process_subcommand_filterset_err;
		}
		// filterset name
		toki++;
		if (strcmp(token[toki], "name")) {
			rc = PFM_ERR_CMD_INVALID;
			goto process_subcommand_filterset_err;
		}
		toki++;
		fname = token[toki];
		fs = fslist_get_by_name(priv_data.filterset_list,
					fname);
		if (!fs) {
			rc = PFM_ERR_NO_ENTRY;
			goto process_subcommand_filterset_err;
		}
		(void)pfm_pnetbuffer_add_rc_string(outdata,
						PFM_ERR_SUCCESS);
		rc = filterset_enumerate_entries(fs,
					list_filter_entry_cb,
					outdata);

	} else {
		rc = PFM_ERR_CMD_INVALID;
		goto process_subcommand_filterset_err;
	}

	return 0;

process_subcommand_filterset_err:
	(void)pfm_pnetbuffer_add_rc_string(outdata, rc);
	return 0;
}


int is_valid_stage_name(char *name)
{
	int		i;
	for (i=0 ; i<PFM_DNS_STAGE_COUNT ; i++) {
		if (!strcmp(stage_string[i], name))
			return 1;
	}
	return 0;
}


int process_subcommand_stage(pfm_t *pfm,
				pnetbuffer_t *outdata,
				char **token,
				int tokenlen)
{
	int		rc;
	int		toki;
	char		*saction;
	char		*sname;
	char		*fname;
	char		**strarray;
	filterset_t	*fs;
	SList		*item;
	SList		*list;
	SList		**listp;

	toki = 0;
	// stage action
	if (!strcmp(token[toki], "attach") ||
	    !strcmp(token[toki], "detach")) {
		if (tokenlen != 5) {
			rc = PFM_ERR_CMD_INVALID;
			goto process_subcommand_stage_err;
		}
		saction = token[toki];
		// stage name
		toki++;
		if (strcmp(token[toki], "name")) {
			rc = PFM_ERR_CMD_INVALID;
			goto process_subcommand_stage_err;
		}
		toki++;
		sname = token[toki];
		if (!is_valid_stage_name(sname)) {
			rc = PFM_ERR_CMD_INVALID;
			goto process_subcommand_stage_err;
		}
		// filterset name
		toki++;
		if (strcmp(token[toki], "filterset")) {
			rc = PFM_ERR_CMD_INVALID;
			goto process_subcommand_stage_err;
		}
		toki++;
		fname = token[toki];

		// validate filterset exists
		fs = fslist_get_by_name(priv_data.filterset_list,
					fname);
		if (!fs) {
			pfm_vector_log(priv_data.module,
				LOG_LEVEL_NOTICE,
				"%s invalid filterset %s",
				__func__,
				fname);
			rc = PFM_ERR_NO_ENTRY;
			goto process_subcommand_stage_err;
		}

		// attach/detach
		if (!strcmp(sname, "query")) {
			list = priv_data.q_filterset_list;
			listp = &priv_data.q_filterset_list;
		} else if (!strcmp(sname, "answer")) {
			list = priv_data.an_filterset_list;
			listp = &priv_data.an_filterset_list;
		} else if (!strcmp(sname, "source")) {
			list = priv_data.src_filterset_list;
			listp = &priv_data.src_filterset_list;
		} else {
			pfm_vector_log(priv_data.module,
				LOG_LEVEL_NOTICE,
				"%s invalid stage %s",
				__func__,
				sname);
			rc = PFM_ERR_CMD_INVALID;
			goto process_subcommand_stage_err;
		}
		if (!strcmp(saction, "attach")) {
			if (fslist_get_by_name(list, fname)) {
				rc = PFM_ERR_CMD_INVALID; // TODO PFM_ERR_EEXISTS
				goto process_subcommand_stage_err;
			}
			rc = slist_item_add(listp, fs);
			if (rc) {
				rc = PFM_ERR_MOD_INTERNAL;
				goto process_subcommand_stage_err;
			}
			pfm_vector_log(priv_data.module,
				LOG_LEVEL_NOTICE,
				"Filterset %s successfully attached to"
				" %s stage",
				fs->name,
				sname);
			(void)pfm_pnetbuffer_add_rc_string(outdata,
							PFM_ERR_SUCCESS);

		} else if (!strcmp(saction, "detach")) {
			if (!fslist_get_by_name(list, fname)) {
				rc = PFM_ERR_NO_ENTRY;
				goto process_subcommand_stage_err;
			}
			rc = slist_item_remove(listp, fs);
			if (rc) {
				rc = PFM_ERR_MOD_INTERNAL;
				goto process_subcommand_stage_err;
			}
			pfm_vector_log(priv_data.module,
				LOG_LEVEL_NOTICE,
				"Filterset %s successfully detached"
				" from stage %s",
				fs->name,
				sname);
			(void)pfm_pnetbuffer_add_rc_string(outdata,
						PFM_ERR_SUCCESS);

		}

	} else if (!strcmp(token[toki], "list")) {
		strarray = calloc(2, sizeof(char*));
		strarray[0] = calloc(256, sizeof(char));
		(void)pfm_pnetbuffer_add_rc_string(outdata,
						PFM_ERR_SUCCESS);
		sname = "query";
		snprintf(strarray[0],
			256,
			"%s filtersets:\n",
			sname);
		(void)pnetbuffer_add_strings(outdata,
				strarray,
				1,
				PNETBUFFER_OPT_STRINGS_COUNT_FIRST);
		item = priv_data.q_filterset_list;
		while (item) {
			fs = slist_get_data(item);
			list_filterset_cb(fs, outdata);
			item = slist_get_next(item);
		}

		sname = "answer";
		snprintf(strarray[0],
			256,
			"\n%s filtersets:\n",
			sname);
		(void)pnetbuffer_add_strings(outdata,
				strarray,
				1,
				PNETBUFFER_OPT_STRINGS_COUNT_FIRST);
		item = priv_data.an_filterset_list;
		while (item) {
			fs = slist_get_data(item);
			list_filterset_cb(fs, outdata);
			item = slist_get_next(item);
		}

		sname = "source";
		snprintf(strarray[0],
			256,
			"\n%s filtersets:\n",
			sname);
		(void)pnetbuffer_add_strings(outdata,
				strarray,
				1,
				PNETBUFFER_OPT_STRINGS_COUNT_FIRST);
		item = priv_data.src_filterset_list;
		while (item) {
			fs = slist_get_data(item);
			list_filterset_cb(fs, outdata);
			item = slist_get_next(item);
		}

		free(strarray[0]);
		free(strarray);

	} else {
		rc = PFM_ERR_CMD_INVALID;
		goto process_subcommand_stage_err;
	}

	return 0;

process_subcommand_stage_err:
	(void)pfm_pnetbuffer_add_rc_string(outdata, rc);
	return 0;
}





int process_subcommand_backend(pfm_t *pfm,
				pnetbuffer_t *outdata,
				char **token,
				int tokenlen)
{
	int			rc;
	int			toki;
	char			*cmd;
	char			*proto;
	char			*host;
	char			*port;
	int			minconns;
	int			maxconns;
	backend_protocol	beproto;

	toki = 0;
	cmd = token[toki];
	if (strcmp(cmd, "add")
	    && strcmp(cmd, "remove") ) {
		pfm_vector_log(priv_data.module,
			LOG_LEVEL_ERROR,
			"%s() expected add or remove command",
			__func__);
		rc = PFM_ERR_CMD_INVALID;
		goto process_subcommand_err;
	}

	if (!strcmp(cmd, "add")) {
		if (tokenlen != 12) {
			pfm_vector_log(priv_data.module,
				LOG_LEVEL_ERROR,
				"%s() wrong number of args for add command",
				__func__);
			rc = PFM_ERR_CMD_INVALID;
			goto process_subcommand_err;
		}
	} else if (!strcmp(cmd, "remove")) {
		if (tokenlen != 7) {
			pfm_vector_log(priv_data.module,
				LOG_LEVEL_ERROR,
				"%s() wrong number of args for remove command",
				__func__);
			rc = PFM_ERR_CMD_INVALID;
			goto process_subcommand_err;
		}
	}

	// protocol
	toki++;
	if (strcmp(token[toki], "proto")) {
		pfm_vector_log(priv_data.module,
			LOG_LEVEL_ERROR,
			"%s() expected proto argument",
			__func__);
		rc = PFM_ERR_CMD_INVALID;
		goto process_subcommand_err;
	}
	toki++;
	proto = token[toki];
	beproto = backend_name_to_protocol(proto);
	if (beproto == _BACKEND_PROTO_MAX) {
		// invalid protocol specified
		pfm_vector_log(priv_data.module,
			LOG_LEVEL_ERROR,
			"%s() invalid proto argument",
			__func__);
		rc = PFM_ERR_CMD_INVALID;
		goto process_subcommand_err;
	}
	// hostname/ip
	toki++;
	if (strcmp(token[toki], "host")) {
		pfm_vector_log(priv_data.module,
			LOG_LEVEL_ERROR,
			"%s() expected host argument",
			__func__);
		rc = PFM_ERR_CMD_INVALID;
		goto process_subcommand_err;
	}
	toki++;
	host = token[toki];
	// port
	toki++;
	if (strcmp(token[toki], "port")) {
		pfm_vector_log(priv_data.module,
			LOG_LEVEL_ERROR,
			"%s() expected port argument",
			__func__);
		rc = PFM_ERR_CMD_INVALID;
		goto process_subcommand_err;
	}
	toki++;
	port = token[toki];

	if (!strcmp(cmd, "add")) {
		// conns
		toki++;
		if (strcmp(token[toki], "conns")) {
			pfm_vector_log(priv_data.module,
				LOG_LEVEL_ERROR,
				"%s() expected conns argument",
				__func__);
			rc = PFM_ERR_CMD_INVALID;
			goto process_subcommand_err;
		}

		// min
		toki++;
		if (strcmp(token[toki], "min")) {
			pfm_vector_log(priv_data.module,
				LOG_LEVEL_ERROR,
				"%s() expected min argument",
				__func__);
			rc = PFM_ERR_CMD_INVALID;
			goto process_subcommand_err;
		}
		toki++;
		minconns = atoi(token[toki]);

		// max
		toki++;
		if (strcmp(token[toki], "max")) {
			pfm_vector_log(priv_data.module,
				LOG_LEVEL_ERROR,
				"%s() expected max argument",
				__func__);
			rc = PFM_ERR_CMD_INVALID;
			goto process_subcommand_err;
		}
		toki++;
		maxconns = atoi(token[toki]);

		rc = backend_add(
			beproto,
			host,
			port,
			minconns,
			maxconns);
		if (rc != 0) {
			pfm_vector_log(priv_data.module,
				LOG_LEVEL_ERROR,
				"%s() Failure during backend_add() rc=%d",
				__func__,
				rc);
//			rc = PFM_ERR_MOD_INTERNAL;
//			goto process_subcommand_err;
		}
		(void)pfm_pnetbuffer_add_rc_string(outdata,
						PFM_ERR_SUCCESS);
	} else if (!strcmp(cmd, "remove")) {
		rc = backend_remove(
			beproto,
			host,
			port);
		if (rc != 0) {
			pfm_vector_log(priv_data.module,
				LOG_LEVEL_ERROR,
				"%s() Failure during backend_remove() rc=%d",
				__func__,
				rc);
//			rc = PFM_ERR_MOD_INTERNAL;
//			goto process_subcommand_err;
		}
		(void)pfm_pnetbuffer_add_rc_string(outdata,
						PFM_ERR_SUCCESS);
	}

	return 0;

process_subcommand_err:
	(void)pfm_pnetbuffer_add_rc_string(outdata, rc);
	return 0;
}




int vector_process_command(pfm_t *pfm,
	pnetbuffer_t *indata,
	pnetbuffer_t *outdata)
{
	int		rc;
	int		i;
	int		count;
	uint8_t		argc;
	char		**argv;
//	char		*str1;
//	char		*str2;
//	char		*proto;
	policy_item_t	*politem;
	
	rc = pnetbuffer_read_strings(indata,
				priv_data.command_array,
				PFM_COMMAND_STR_SIZE,
				PFM_COMMAND_TOKEN_MAX,
				&argc,
				PNETBUFFER_OPT_STRINGS_COUNT_FIRST);
	if (rc != 0) {
		(void)pfm_pnetbuffer_add_rc_string(outdata,
					PFM_ERR_CMD_INVALID);
		return rc;
	}
	if (argc < 1) {
		(void)pfm_pnetbuffer_add_rc_string(outdata,
					PFM_ERR_CMD_INVALID);
		return 0;
	}
	argv = priv_data.command_array;
	i = 0;

	if (!strcmp(argv[i], "counters")) {
		i++;
		if (!strcmp(argv[i], "show")) {
			(void)pfm_pnetbuffer_add_rc_string(outdata,
						PFM_ERR_SUCCESS);
			for (count=0 ; count<_METRIC_COUNT ; count++) {
				snprintf (priv_data.response_array[0],
					RESPONSE_STRING_SIZE,
					"%s %d\n",
					metric_get_name(
						&priv_data.metric[count]),
					metric_get_u32(
						&priv_data.metric[count]));
				rc = pnetbuffer_add_strings(outdata,
						priv_data.response_array,
						1,
						PNETBUFFER_OPT_STRINGS_COUNT_FIRST);
			}
			if (rc != 0) {
				pnetbuffer_reset(outdata);
				(void)pfm_pnetbuffer_add_rc_string(outdata,
							PFM_ERR_MOD_INTERNAL);
				return 0;
			}
		} else if (!strcmp(argv[i], "reset")) {
			metric_set_u32(
				&priv_data.metric[METRIC_QUERY_TOTAL],
				0);
			metric_set_u32(
				&priv_data.metric[METRIC_QUERY_FILTERED],
				0);
			metric_set_u32(
				&priv_data.metric[METRIC_ANSWER_TOTAL],
				0);
			metric_set_u32(
				&priv_data.metric[METRIC_ANSWER_FILTERED],
				0);
			(void)pfm_pnetbuffer_add_rc_string(outdata,
				PFM_ERR_SUCCESS);
		} else {
			(void)pfm_pnetbuffer_add_rc_string(outdata,
				PFM_ERR_CMD_INVALID);
			return 0;
		}
		return 0;

	} else if (!strcmp(argv[i], "backend")) {
		i++;
		rc = process_subcommand_backend(pfm,
						outdata,
						&argv[i],
						argc-i);
		pfm_vector_log(priv_data.module,
			LOG_LEVEL_ERROR,
			"%s() process_subcommand_backend() rc=%d",
			__func__,
			rc);

	} else if (!strcmp(argv[i], "policy")) {
		i++;
		if (!strcmp(argv[i], "list")) {
			(void)pfm_pnetbuffer_add_rc_string(outdata, PFM_ERR_SUCCESS);
			i = 0;
			while (policy_table[i].name != NULL) {
				if (priv_data.policy_bits & (1<<policy_table[i].bit)) {
					snprintf (priv_data.response_array[0],
						RESPONSE_STRING_SIZE,
						"policy set %s\n",
						policy_table[i].name);
					rc = pnetbuffer_add_strings(outdata,
								priv_data.response_array,
								1,
								PNETBUFFER_OPT_STRINGS_COUNT_FIRST);
					if (rc != 0) {
						pnetbuffer_reset(outdata);
						(void)pfm_pnetbuffer_add_rc_string(outdata,
								PFM_ERR_MOD_INTERNAL);
						return 0;
					}
				}
				i++;
			}
			return 0;
		} else if (!strcmp(argv[i], "set")) {
			i++;
			politem = policy_item_lookup_by_name(argv[i]);
			if (!politem) {
				(void)pfm_pnetbuffer_add_rc_string(outdata,
								PFM_ERR_CMD_INVALID);
				return 0;
			}
			priv_data.policy_bits |= (1<<politem->bit);
			(void)pfm_pnetbuffer_add_rc_string(outdata, PFM_ERR_SUCCESS);
			return 0;
		} else if (!strcmp(argv[i], "unset")) {
			i++;
			politem = policy_item_lookup_by_name(argv[i]);
			if (!politem) {
				(void)pfm_pnetbuffer_add_rc_string(outdata,
								PFM_ERR_CMD_INVALID);
				return 0;
			}
			priv_data.policy_bits &= ~(1<<politem->bit);
			(void)pfm_pnetbuffer_add_rc_string(outdata, PFM_ERR_SUCCESS);
			return 0;
		} else {
			(void)pfm_pnetbuffer_add_rc_string(outdata,
						PFM_ERR_CMD_INVALID);
			return 0;
		}

	} else if (!strcmp(argv[i], "filterset")) {
		i++;
		rc = process_subcommand_filterset(pfm,
						outdata,
						&argv[i],
						argc-i);

	} else if (!strcmp(argv[i], "stage")) {
		i++;
		rc = process_subcommand_stage(pfm,
						outdata,
						&argv[i],
						argc-i);

	} else if (!strcmp(argv[i], "filter")) {
		pfm_vector_log(priv_data.module,
			LOG_LEVEL_WARNING,
			"Use of filter command is deprecated and"
			" will be ignored, convert this to a filterset");
			(void)pfm_pnetbuffer_add_rc_string(outdata,
						PFM_ERR_CMD_INVALID);
		return 0;

	} else {
		(void)pfm_pnetbuffer_add_rc_string(outdata, PFM_ERR_CMD_INVALID);
		return 0;
	}
	return 0;
}


int pnetbuffer_sendto(pnetbuffer_t *nb,
			int fd,
			int flags,
			struct sockaddr *sa,
			socklen_t salen)
{
	int		rc;
	uint16_t	plen;
	char		*p;

	pnetbuffer_get_readable(nb, &p, &plen);
	rc = sendto(fd, p, plen, flags, sa, salen);
	return rc;
}


int dns_packet_read_header(pnetbuffer_t *indata,
			uint16_t *id,
			uint16_t *flags,
			uint16_t *q_count,
			uint16_t *a_count,
			uint16_t *ns_count,
			uint16_t *ar_count)
{
	int		rc;
	rc = pnetbuffer_read_uint16(indata, id);
	if (rc != 0) return rc;
	rc = pnetbuffer_read_uint16(indata, flags);
	if (rc != 0) return rc;
	rc = pnetbuffer_read_uint16(indata, q_count);
	if (rc != 0) return rc;
	rc = pnetbuffer_read_uint16(indata, a_count);
	if (rc != 0) return rc;
	rc = pnetbuffer_read_uint16(indata, ns_count);
	if (rc != 0) return rc;
	rc = pnetbuffer_read_uint16(indata, ar_count);
	if (rc != 0) return rc;
	return 0;
}


int dns_packet_write_header(pnetbuffer_t *nb,
			uint16_t id,
			uint16_t flags,
			uint16_t q_count,
			uint16_t a_count,
			uint16_t ns_count,
			uint16_t ar_count)
{
	int		rc;
	rc = pnetbuffer_add_uint16(nb, id);
	if (rc != 0) return rc;
	rc = pnetbuffer_add_uint16(nb, flags);
	if (rc != 0) return rc;
	rc = pnetbuffer_add_uint16(nb, q_count);
	if (rc != 0) return rc;
	rc = pnetbuffer_add_uint16(nb, a_count);
	if (rc != 0) return rc;
	rc = pnetbuffer_add_uint16(nb, ns_count);
	if (rc != 0) return rc;
	rc = pnetbuffer_add_uint16(nb, ar_count);
	if (rc != 0) return rc;
//	uint16_t plen;
//	char	*p;
//	pnetbuffer_get_readable(nb, &p, &plen);
	return 0;
}


packetbuf_t *packetbuf_get_or_alloc(void)
{
	dlist_t			*item;
	packetbuf_t		*pb;

	// get an existing item from the free list	
	if (!dlist_is_empty(&priv_data.packetbuf_free_head)) {
		item = (&priv_data.packetbuf_free_head)->next;
		pfm_vector_log(priv_data.module,
			LOG_LEVEL_DEBUG,
			"%s() claiming from freelist pb=%p ->prev=%p ->next=%p",
			__func__,
			item,
			item->prev,
			item->next);
		dlist_remove(item);
		item->next = NULL;
		item->prev = NULL;
		pb = (packetbuf_t*)item;
		return pb;
	}

	// or allocate a new item
	pb = calloc(1, sizeof(packetbuf_t));
	if (! pb)
		return NULL;
	pb->nb = pnetbuffer_init(pb->nb_buffer, PACKET_BUFFER_SIZE);
	pfm_vector_log(priv_data.module,
		LOG_LEVEL_DEBUG,
		"%s() allocated new pb=%p",
		__func__,
		pb);

	return pb;
}


void packetbuf_free(packetbuf_t *pb)
{
	pfm_vector_log(priv_data.module,
		LOG_LEVEL_DEBUG,
		"%s() returning item=%p",
		__func__,
		(dlist_t*)pb);
	pnetbuffer_reset(pb->nb);
	dlist_add(&priv_data.packetbuf_free_head, (dlist_t*)pb);
	return;
}


int dns_fqdn_write_netbuffer(pnetbuffer_t *nb, char *str)
{
	int		rc;
	uint8_t		len;
	char		*p;
	char		*pp;
	p = str;
	while (1) {
		pp = strstr(p, ".");
		if (pp) {
			if (pp == p) {
				p++;
				continue;
			}
			len = pp - p;
		} else {
			len = strlen(p);
		}
		rc = pnetbuffer_add_uint8(nb, len);
		if (rc != 0) return rc;
		rc = pnetbuffer_add_data(nb, p, len);
		if (rc != 0) return rc;
		if (!pp) break; // end of string
		p = pp++;
	}

	return 0;

}


int packetbuf_try_read_from_fd(packetbuf_t *pb,
			int fd,
			int *read_len)
{
	int		rc;
	int		rl;
	uint16_t	plen;
	char		*p;

	pb->saddr_len = sizeof(struct sockaddr);
	pnetbuffer_get_writable(pb->nb, &p, &plen);
	fflush(stdout);
	assert(plen > 0);
	rl = recvfrom(fd,
			p,
			(int)plen,
			0,
			&pb->saddr,
			&pb->saddr_len);
	if (rl == -1) {
		rc = errno;
		*read_len = 0;
		pfm_vector_log(priv_data.module,
			LOG_LEVEL_WARNING,
			"Error %d occurred during recvfrom()"
			" TODO close client connection",
			rc);
		return rc;
	}
	if (read_len == 0) {
		rc = errno;
		*read_len = 0;
		pfm_vector_log(priv_data.module,
			LOG_LEVEL_DEBUG,
			"%s() Zero bytes read in recvfrom()"
			" TODO close client connection",
			__func__,
			rc);
		return rc;
	}
	*read_len = rl;
	rc = pnetbuffer_set_written(pb->nb, (uint16_t)rl);
	assert(rc == 0);
	pfm_vector_log(priv_data.module,
		LOG_LEVEL_DEBUG,
		"%s() recvfrom() fd=%d bytes=%d pb->saddr_len=%d",
		__func__,
		fd,
		rl,
		pb->saddr_len);
	return 0;
}



int tcp_try_read_packet(int fd, packetbuf_t **rpb)
{
	int		rc;
	int		read_len;
	uint16_t	plen;
	fd_t		*fdte;

	*rpb = NULL;
	fdte = &priv_data.fd_table[fd];
	if (!fdte->in_pb) {
		pfm_vector_log(priv_data.module,
			LOG_LEVEL_DEBUG,
			"%s() allocating new in_pb",
			__func__);
		// start reading a new packet
		fdte->in_pb = packetbuf_get_or_alloc();
		assert(fdte->in_pb);
		fdte->in_rem = 0;
	}
	if (fdte->usagetype == DNS_USAGE_TYPE_FRONTEND)
		fdte->in_pb->sfd = fd;
//	dump_fdte(fdte);

	rc = packetbuf_try_read_from_fd(fdte->in_pb,
					fd,
					&read_len);

	if (rc && rc != EAGAIN && rc != EWOULDBLOCK) {
		// something went wrong
		pfm_vector_log(priv_data.module,
			LOG_LEVEL_WARNING,
			"Error %d occurred during read",
			rc);
		packetbuf_free(fdte->in_pb);
		fdte->in_pb = NULL;
		fdte->in_rem = 0;
		return rc;
	}
	if (!read_len) {
		// no error and no bytes, socket was closed
		pfm_vector_log(priv_data.module,
			LOG_LEVEL_DEBUG,
			"Zero bytes read in recvfrom()");
		packetbuf_free(fdte->in_pb);
		fdte->in_pb = NULL;
		fdte->in_rem = 0;
		return ESHUTDOWN;
	}

	if (!fdte->in_rem) {
		// Read packet length header (TCP only)
		rc = pnetbuffer_read_uint16(fdte->in_pb->nb, &plen);
		assert (rc == 0);
		pfm_vector_log(priv_data.module,
			LOG_LEVEL_DEBUG,
			"%s() DNS request length is %d bytes",
			__func__,
			plen);
		if (read_len < plen) {
			// still need more data
			fdte->in_rem = plen - read_len;
			return EAGAIN;
		}
	} else {
		// packet data continuation
		fdte->in_rem -= read_len;
		if (fdte->in_rem)
			return EAGAIN; // more to read
	}

	// return the complete packet
//	dump_fdte(fdte);
	assert(fdte->in_pb);
	*rpb = fdte->in_pb;
	return 0;
}


int fd_try_writes(int fd,
		struct sockaddr *dest_sa)
{
	int			rc;
	uint16_t		len;
	socklen_t		salen;
	char			*p;
	fd_t			*fdte;
	packetbuf_t		*pb;
	ioctl_opt_modify_fd_t	opt;

	assert (fd > 0);
	fdte = &priv_data.fd_table[fd];
	if (dest_sa)
		salen = sizeof(struct sockaddr);
	else
		salen = 0;

	if (fdte->out_pb) {
		// continue sending current packet
		pb = fdte->out_pb;
		pfm_vector_log(priv_data.module,
			LOG_LEVEL_DEBUG,
			"%s() continue sending current packet fd=%d",
			__func__,
			fd);
	} else if (!dlist_is_empty(&fdte->out_qhead)) {
		// start sending the next queued packet
		fdte->out_pb = (packetbuf_t*)(&fdte->out_qhead)->next;
		pb = fdte->out_pb;
		dlist_remove((dlist_t *)pb);
		pfm_vector_log(priv_data.module,
			LOG_LEVEL_DEBUG,
			"%s() start sending next packet fd=%d",
			__func__,
			fd);
	} else {
		// nothing to send, disable write events
		opt.fd = fd;
		opt.flags = 0;
		rc = pfm_vector_ioctl(priv_data.module,
				PFM_IOCTL_MODIFY_BACKEND_FD,
				&opt);
		pfm_vector_log(priv_data.module,
			LOG_LEVEL_DEBUG,
			"%s() nothing to send, disable write events fd=%d",
			__func__,
			fd);
		return 0;
	}

	// send as much data as possible
	while (1) {
		pnetbuffer_get_readable(pb->nb, &p, &len);
		rc = sendto(fd,
			p,
			len,
			0,
			dest_sa,
			salen);
		pfm_vector_log(priv_data.module,
			LOG_LEVEL_DEBUG,
			"%s() sendto() fd=%d len=%d rc=%d pb=%p",
			__func__,
			fd,
			len,
			rc,
			pb);
		if (rc == -1) {
			rc = errno;
			if (rc == EAGAIN || rc == EWOULDBLOCK) {
				fdte->out_pb = pb;
				opt.fd = fd;
				opt.flags = PFM_FD_MODIFY_F_WRITABLE;
				rc = pfm_vector_ioctl(priv_data.module,
						PFM_IOCTL_MODIFY_BACKEND_FD,
						&opt);
				return EAGAIN;
			}
			pfm_vector_log(priv_data.module,
				LOG_LEVEL_ERROR,
				"%s() sendto() rc=%d fd=%d",
				__func__,
				rc,
				fd);
			return rc;
		} else if (rc > 0) {
			pfm_vector_log(priv_data.module,
				LOG_LEVEL_DEBUG,
				"%s() sendto() wrote %d bytes to fd=%d",
				__func__,
				rc,
				fd);
			if (rc == len) {
				// entire packet written, return to free list
				pfm_vector_log(priv_data.module,
					LOG_LEVEL_DEBUG,
					"%s() entire packet written, return to free list",
					__func__);
				pnetbuffer_reset(pb->nb);
//				dlist_add(&priv_data.packetbuf_free_head,
//					(dlist_t *)pb);
				// if there are queued packets keep going
				if (!dlist_is_empty(&fdte->out_qhead)) {
					packetbuf_free(pb); // free prior pb
					pb = (packetbuf_t *)fdte->out_qhead.next;
					dlist_remove((dlist_t *)pb);
					fdte->out_pb = pb;
					continue;
				}
				// no more data, disable writeable events
				if (fdte->usagetype == DNS_USAGE_TYPE_FRONTEND)
					request_reset(pb->req);
				packetbuf_free(pb); // free pb after resetting request object
				pb = NULL;
				fdte->out_pb = NULL;
				opt.fd = fd;
				opt.flags = 0;
				rc = pfm_vector_ioctl(priv_data.module,
						PFM_IOCTL_MODIFY_BACKEND_FD,
						&opt);
				return 0;
			} else {
				// a portion was written, advance the buffer
				// and try the write again
				pfm_vector_log(priv_data.module,
					LOG_LEVEL_DEBUG,
					"%s() a portion was written, advance the buffer",
					__func__);
				pnetbuffer_read_advance(pb->nb, rc);
				continue;
			}
		}
	} // while

	return 0;
}


int fd_write_or_queue_packetbuf(int fd,
			packetbuf_t *pb,
			struct sockaddr *dest_sa)
{
	int			rc;
//	uint16_t		len;
//	char			*p;
	fd_t			*fdte;
	ioctl_opt_modify_fd_t	opt;

	fdte = &priv_data.fd_table[fd];
	pfm_vector_log(priv_data.module,
		LOG_LEVEL_DEBUG,
		"%s() fd=%d pb=%p",
		__func__,
		fd,
		pb);
//	dump_packetbuf(pb);
//	dump_fdte(fdte);

	if (fdte->out_pb) {
		// already sending a packet, so queue this one
		// and enable writable events
		dlist_add_tail(&fdte->out_qhead, (dlist_t*)pb);
		pfm_vector_log(priv_data.module,
			LOG_LEVEL_DEBUG,
			"%s() fd=%d currently sending, queuing packet",
			__func__,
			fd);
//		opt.fd = fd;
//		opt.flags = PFM_FD_MODIFY_F_WRITABLE;
//		rc = pfm_vector_ioctl(priv_data.module,
//				PFM_IOCTL_MODIFY_BACKEND_FD,
//				&opt);
//		return 0;
	} else {
		fdte->out_pb = pb;
	}

	// send as much data as possible without blocking
	rc = fd_try_writes(fd, dest_sa);
	if (rc == EAGAIN) {
		// ensure write events are enabled for this fd
		opt.fd = fd;
		opt.flags = PFM_FD_MODIFY_F_WRITABLE;
		rc = pfm_vector_ioctl(priv_data.module,
				PFM_IOCTL_MODIFY_BACKEND_FD,
				&opt);
		return 0;
	} else if (rc) {
		// TODO some other error on the socket
		pfm_vector_log(priv_data.module,
			LOG_LEVEL_ERROR,
			"%s() fd=%d rc=%d",
			__func__,
			fd,
			rc);
//		fprintf(stderr,"%s() fd=%d rc=%d\n",
//			__func__,
//			fd,
//			rc);
		fflush(stderr);
		fflush(stdout);
		//assert(rc == 0);
	}

	return 0;
}


int send_dns_error_response(int fd,
			struct sockaddr *dest_addr,
			uint16_t qcount,
			uint16_t flags,
			uint16_t id,
			char *fqdn,
			uint16_t rtype,
			uint16_t rclass,
			uint16_t err_code)
{
	int		rc;
	uint16_t	u16;
	char		*p;
	packetbuf_t	*pb;
	packetbuf_t	*pbwrap;

	pfm_vector_log(priv_data.module,
		LOG_LEVEL_DEBUG,
		"%s() fd=%d id=%d fqdn=%s rtype=%d rclass=%d err_code=%d",
		__func__,
		fd,
		id,
		fqdn,
		rtype,
		rclass,
		err_code);
	pb = packetbuf_get_or_alloc();
	assert(pb != NULL);

	u16 = DNS_F_QR|DNS_F_AA|err_code;
	if (flags & DNS_F_RD)
		u16 |= (DNS_F_RD|DNS_F_RA);
	pnetbuffer_reset(pb->nb);
	rc = dns_packet_write_header(pb->nb,
				id,
				u16,
				qcount,
				0,
				0,
				0);
	assert(rc == 0);
	if (rc != 0) return rc;
	rc = dns_fqdn_write_netbuffer(pb->nb, fqdn);
	assert(rc == 0);
	if (rc != 0) return rc;
	rc = pnetbuffer_add_uint16(pb->nb, rtype);
	assert(rc == 0);
	if (rc != 0) return rc;
	rc = pnetbuffer_add_uint16(pb->nb, rclass);
	assert(rc == 0);
	if (rc != 0) return rc;

//	pnetbuffer_get_readable(pb->nb, &p, &u16);
//	dump_memory_hex(p, u16);

	dump_fdte(&priv_data.fd_table[fd]);
	if (priv_data.fd_table[fd].conntype == DNS_CONN_TYPE_TCP) {
		// prepend the TCP length header using a 2nd packetbuf
		pnetbuffer_get_readable(pb->nb, &p, &u16);
//		dump_memory_hex(p, u16);
		pfm_vector_log(priv_data.module,
			LOG_LEVEL_DEBUG,
			"%s() netbuffer pre-wrap fd=%d pb=%p pb->nb=%p len(pb->nb)=%d",
			__func__,
			fd,
			pb,
			pb->nb,
			u16);
		assert(u16 > 0);

		pbwrap = packetbuf_get_or_alloc();
		assert(pbwrap != NULL);
		assert(pbwrap != pb);
		pnetbuffer_reset(pbwrap->nb);

		rc = pnetbuffer_add_uint16(pbwrap->nb, u16);
		assert(rc == 0);
		rc = pnetbuffer_add_data(pbwrap->nb, p, u16);
		assert(rc == 0);

		pnetbuffer_get_readable(pbwrap->nb, &p, &u16);
		pfm_vector_log(priv_data.module,
			LOG_LEVEL_DEBUG,
			"%s() netbuffer post-wrap fd=%d pbwrap=%p pbwrap->nb=%p len(pbwrap->nb)=%d",
			__func__,
			fd,
			pbwrap,
			pbwrap->nb,
			u16);
		packetbuf_free(pb);
		pb = pbwrap;
		pnetbuffer_get_readable(pb->nb, &p, &u16);
		pfm_vector_log(priv_data.module,
			LOG_LEVEL_DEBUG,
			"%s() netbuffer post-wrap fd=%d pb=%p pb->nb=%p len(pb->nb)=%d",
			__func__,
			fd,
			pb,
			pb->nb,
			u16);
	}

//	pnetbuffer_get_readable(pb->nb, &p, &u16);
//	dump_memory_hex(p, u16);
//	pfm_vector_log(priv_data.module,
//		LOG_LEVEL_DEBUG,
//		"%s() fd=%d pb=%p pb->nb=%p len(pb->nb)=%d",
//		__func__,
//		fd,
//		pb,
//		pb->nb,
//		u16);

	assert(fd > 0);
	rc = fd_write_or_queue_packetbuf(fd, pb, dest_addr);
	if (rc != 0) {
		pfm_vector_log(priv_data.module,
			LOG_LEVEL_WARNING,
			"Failed to send error response"
			"fd=%d source=%s name='%s' err=%d",
			fd,
			inet_ntoa(((struct sockaddr_in*)dest_addr)->sin_addr),
			fqdn,
			err_code);
	}
	return rc;
}


backend_conn_err_t select_usable_backend_fd(dns_conn_type protocol,
			uint16_t	id,
			backend_t	**pbe,
			int		*pbe_fd,
			fd_t		**pfdte,
			request_t	**preq)
{
	int		rc;
	int		be_i;
	int		bc_i;
	int		be_fd;
	backend_t	*backend;
	fd_t		*fdte;
	request_t	*req;

	if (priv_data.backend_count[protocol] == 0)
		return RC_BACKEND_NO_PROTO;

	be_i = (random() % priv_data.backend_count[protocol]);
	backend = priv_data.backend[protocol][be_i];
	assert(backend);
	while (1) {
		if (backend->conn_count == 0) {
			// schedule a connection attempt which provides a
			// backend connection(stub) to queue the request on
			rc = backend_add_connection(backend);
			(void)rc;
			if (be_i > 0) {
				// try another backend
				pfm_vector_log(priv_data.module,
					LOG_LEVEL_NOTICE,
					"%s() backend=%p be_i=%d trying another backend",
					__func__,
					backend,
					be_i);
				be_i--;
				backend = priv_data.backend[protocol][be_i];
				assert(backend);
				continue;
			}
			// no more backends to try
			goto select_usable_backend_fd_err;
		}
		bc_i = (random() % backend->conn_count);
		while (1) {
			be_fd = backend->conn_fd_table[bc_i];
			fdte = &priv_data.fd_table[be_fd];
			assert(fdte);
			req = request_table_get_entry(fdte->ptr.request_table, id);
			assert(req);
			if (req->frontend_fd <= 0 && req->backend_fd <= 0) {
				// use this backend/fd
				backend_count_usage(backend);
				*pbe = backend;
				*pbe_fd = be_fd;
				*pfdte = fdte;
				*preq = req;
				return 0;
			}
			// ID already inuse, pick another backend connection
			if (bc_i > 0) {
				pfm_vector_log(priv_data.module,
					LOG_LEVEL_NOTICE,
					"%s() backend=%p bc_i=%d trying another backend connection ",
					__func__,
					backend,
					bc_i);
				bc_i--;
				continue;
			}
			// no more FDs in this backend, try another backend
			if (be_i > 0) {
				// try another backend
				pfm_vector_log(priv_data.module,
					LOG_LEVEL_NOTICE,
					"%s() backend=%p be_i=%d trying another backend",
					__func__,
					backend,
					be_i);
				be_i--;
				backend = priv_data.backend[protocol][be_i];
				assert(backend);
				break;
			}
			// nothing else to try
			goto select_usable_backend_fd_err;
		}
	}
select_usable_backend_fd_err:
	return RC_BACKEND_NO_CONNS;
}



int vector_process_data(pfm_t *pfm,
			packetbuf_t *in_pb)
{
	int			rc;
	int			i;
//	int			conn_id;
//	int			backend_id;
	int			be_fd;
	uint16_t		id;
	uint16_t		flags;
	uint16_t		q_count;
	uint16_t		a_count;
	uint16_t		ns_count;
	uint16_t		ar_count;
	uint16_t		type;
	uint16_t		class;
	uint16_t		u16;
	uint8_t			len;
	dns_conn_type		protocol;
	backend_conn_err_t	be_rc;
	char			str[1024];
	char			*p;
	SList			*item;
	filter_t		*filter;
	filterset_t		*fs;
	request_t		*req;
	backend_t		*backend;
	fd_t			*fdte;
	struct sockaddr_in	*sin;
	pnetbuffer_t		*indata;

	indata = in_pb->nb;
	type = 0;
	class = 0;

//	// debugging only
//	dump_packetbuf(in_pb);
	pnetbuffer_get_readable(indata, &p, &u16);
//	dump_memory_hex(p, u16);

	assert(in_pb->sfd > 0);
	protocol = priv_data.fd_table[in_pb->sfd].conntype;
	pfm_vector_log(priv_data.module,
			LOG_LEVEL_DEBUG,
			"%s() in_pb->sfd=%d protocol=%d len=%d",
			__func__,
			in_pb->sfd,
			protocol,
			u16);

	rc = dns_packet_read_header(indata,
				&id,
				&flags,
				&q_count,
				&a_count,
				&ns_count,
				&ar_count);
	if (rc != 0) return rc; //malformed
	assert(id != 0);
	pfm_vector_log(priv_data.module,
			LOG_LEVEL_INFO,
			"%s() DNS request id=%d flags=%u q_count=%d"
			" a_count=%d ns_count=%d ar_count=%d",
			__func__,
			id,
			flags,
			q_count,
			a_count,
			ns_count,
			ar_count);
		
	if (priv_data.policy_bits & POLICY_DENY_MULTIPLE_QUERIES) {
		sin = (struct sockaddr_in *)&in_pb->saddr;
		pfm_vector_log(priv_data.module,
				LOG_LEVEL_WARNING,
				"WARN dns request denied by policy: multiple "
				"queries present proto=%s source=%s q_count=%d"
				" type=%d",
				backend_protocol_to_name(protocol),
				inet_ntoa(sin->sin_addr),
				q_count,
				type);
		metric_incr_u32(&priv_data.metric[METRIC_QUERY_FILTERED]);
		return VOTE_DENY;
	}

	// process query section
	filter = NULL;
	for (i=0 ; i<q_count ; i++) {
		p = str;
		*p = 0;
		while (1) {
			rc = pnetbuffer_read_uint8(indata, &len);
			if (rc != 0) return VOTE_DROP; //malformed
			if (len == 0) break;
			rc = pnetbuffer_read_data(indata, p, len);
			if (rc != 0) return VOTE_DROP; //malformed
			p += len;
			*p = '.';
			p++;
		}
		*p = 0;
		rc = pnetbuffer_read_uint16(indata, &type);
		if (rc != 0) return VOTE_DROP; //malformed
		rc = pnetbuffer_read_uint16(indata, &class);
		if (rc != 0) return VOTE_DROP; //malformed
		metric_incr_u32(&priv_data.metric[METRIC_QUERY_TOTAL]);

		// filter query
		item = priv_data.q_filterset_list;
		while (item) {
			fs = slist_get_data(item);
			filter = filterset_match(fs, str);
			if (filter) {
				break;
			}
			item = slist_get_next(item);
		}
		if (!item) {
			// no match found
			filter = NULL;
		}
		if (filter) {
			if (filter->action == F_ACTION_REJECT) {
				sin = (struct sockaddr_in *)&in_pb->saddr;
				pfm_vector_log(priv_data.module,
					LOG_LEVEL_WARNING,
					"dns query denied proto=%s source=%s"
					" name='%s' "
					"type=%d class=%d id=%d",
					backend_protocol_to_name(protocol),
					inet_ntoa(sin->sin_addr),
					str,
					type,
					class,
					id);
				metric_incr_u32(&priv_data.metric[METRIC_QUERY_FILTERED]);

				// send NXDOMAIN response
				rc = send_dns_error_response(in_pb->sfd,
					&in_pb->saddr,
					q_count,
					flags,
					id,
					str,
					type,
					class,
					DNS_F_RC_NXDOMAIN);
				if (rc != 0) {
					pfm_vector_log(priv_data.module,
						LOG_LEVEL_WARNING,
						"Failed to send nxdomain "
						"response source=%s name='%s'",
						inet_ntoa(sin->sin_addr),
						str);
				}
				return VOTE_DENY; //io error

			} else if (filter->action == F_ACTION_LOG) {
				sin = (struct sockaddr_in *)&in_pb->saddr;
				pfm_vector_log(priv_data.module,
					LOG_LEVEL_NOTICE,
					"dns query logged"
					" proto=%s source=%s "
					"name='%s' type=%d class=%d id=%d",
					backend_protocol_to_name(protocol),
					inet_ntoa(sin->sin_addr),
					str,
					type,
					class,
					id);
			}
		}

	}

	if (!filter || filter->action == F_ACTION_ACCEPT) {
		// log 'pass' queries at info level or more
		sin = (struct sockaddr_in *)&in_pb->saddr;
		pfm_vector_log(priv_data.module,
			LOG_LEVEL_INFO,
			"dns query allowed proto=%s source=%s name='%s'"
			" type=%d class=%d id=%d",
			backend_protocol_to_name(protocol),
			inet_ntoa(sin->sin_addr),
			str,
			type,
			class,
			id);
	}


/*
	if (priv_data.backend_count[protocol] == 0) {
		pfm_vector_log(priv_data.module,
			LOG_LEVEL_WARNING,
			"No backend configured for %s, dropping packet",
			backend_protocol_to_name(protocol));
		rc = send_dns_error_response(in_pb->sfd,
					&in_pb->saddr,
					q_count,
					flags,
					id,
					str,
					type,
					class,
					DNS_F_RC_SRVERROR);
		return 0;
	}

	// select backend and connection to use
	backend_id = (random() % priv_data.backend_count[protocol]);
	backend = priv_data.backend[protocol][backend_id];
	backend_count_usage(backend);
	if (backend->conn_count == 0) {
		// schedule a connection attempt which provides a
		// backend connection(stub) to queue the request on
		rc = backend_add_connection(backend);
		if (rc) {
			pfm_vector_log(priv_data.module,
				LOG_LEVEL_WARNING,
				"Backend has no connections for %s, dropping packet",
				backend_protocol_to_name(protocol));
			rc = send_dns_error_response(in_pb->sfd,
						&in_pb->saddr,
						q_count,
						flags,
						id,
						str,
						type,
						class,
						DNS_F_RC_SRVERROR);
			return 0;
		}
	}

	conn_id = (random() % backend->conn_count);
	be_fd = backend->conn_fd_table[conn_id];
	pfm_vector_log(priv_data.module,
		LOG_LEVEL_DEBUG,
		"%s() Selected backend_id=%d backend=%p conn_id=%d fd=%d",
		__func__,
		backend_id,
		backend,
		conn_id,
		be_fd);
	if (be_fd == -1) {
		pfm_vector_log(priv_data.module,
			LOG_LEVEL_EMERG,
			"%s() invalid FD in backend connection table,"
			" exiting! backend=%p index=%d",
			__func__,
			backend,
			conn_id);
		exit(EINVAL);
	}
	fdte = &priv_data.fd_table[be_fd];
	pfm_vector_log(priv_data.module,
		LOG_LEVEL_DEBUG,
		"%s() backend=%p protocol=%s fd=%d fdte=%p",
		__func__,
		backend,
		backend_protocol_to_name(protocol),
		be_fd,
		fdte);
	pfm_vector_log(priv_data.module,
		LOG_LEVEL_DEBUG,
		"%s() fdte->ptr.ptr=%p",
		__func__,
		fdte->ptr.ptr);

	// setup request table entry
	req = request_table_get_entry(fdte->ptr.request_table, id);
	if (!req) {
		pfm_vector_log(priv_data.module,
			LOG_LEVEL_WARNING,
			"%s() could not retrieve request from table, dropping request"
			" in_pb->sfd=%d protocol=%d id=%d",
			__func__,
			in_pb->sfd,
			protocol,
			id);
		fflush(stdout);
		assert(0);
		return VOTE_DROP;
	}
*/

	be_rc = select_usable_backend_fd(protocol, id, &backend, &be_fd, &fdte, &req);
	if (be_rc == RC_BACKEND_NO_PROTO) {
		pfm_vector_log(priv_data.module,
			LOG_LEVEL_WARNING,
			"No backend configured for %s, dropping packet",
			backend_protocol_to_name(protocol));
		rc = send_dns_error_response(in_pb->sfd,
					&in_pb->saddr,
					q_count,
					flags,
					id,
					str,
					type,
					class,
					DNS_F_RC_SRVERROR);
//		return 0;
		return VOTE_DROP;
	} else if (be_rc == RC_BACKEND_NO_CONNS) {
		pfm_vector_log(priv_data.module,
			LOG_LEVEL_WARNING,
			"Backend has no connections for %s, dropping packet",
			backend_protocol_to_name(protocol));
		rc = send_dns_error_response(in_pb->sfd,
					&in_pb->saddr,
					q_count,
					flags,
					id,
					str,
					type,
					class,
					DNS_F_RC_SRVERROR);
//		return 0;
		return VOTE_DROP;
	}
	pfm_vector_log(priv_data.module,
		LOG_LEVEL_DEBUG,
		"%s() Selected backend=%p fd=%d req=%p",
		__func__,
		backend,
		be_fd,
		req);

	// recheck... was already done in select_usable_backend_fd() above 
	if (!(req->frontend_fd <= 0)) {
		pfm_vector_log(priv_data.module,
			LOG_LEVEL_WARNING,
			"%s() request table entry frontend_fd is not unset,"
			" continuing anyway (reset needed?)"
			" in_pb->sfd=%d id=%d req=%p .frontend_fd=%d .backend_fd=%d .id=%d",
			__func__,
			in_pb->sfd,
			id,
			req,
			req->frontend_fd,
			req->backend_fd,
			req->req_id
			);
//			dump_request(req);
		fflush(stdout);
		assert(req->frontend_fd <= 0);
	}
//	assert(req->backend_fd <= 0);
	req->src_sockaddr = in_pb->saddr;
	req->req_id = id;
	req->frontend_fd = in_pb->sfd;
	req->backend_fd = be_fd;
	(void)clock_gettime(CLOCK_MONOTONIC, &req->start_ts);
	in_pb->req = req;

	fflush(stdout);
	// forward request to backend for processing
	pnetbuffer_reset_read(in_pb->nb); // include TCP length header
	assert(req->backend_fd > 0);
	rc = fd_write_or_queue_packetbuf(req->backend_fd,
					in_pb,
					&backend->dest_sa);
	if (rc != 0) {
		pfm_vector_log(priv_data.module,
			LOG_LEVEL_WARNING,
			"Failed to send backend request"
			"fd=%d source=%s err=%d",
			req->backend_fd,
			inet_ntoa(((struct sockaddr_in*)&backend->dest_sa)->sin_addr),
			rc);
			request_reset(req);
	}

	pfm_vector_log(priv_data.module,
		LOG_LEVEL_DEBUG,
		"%s() sent packet to backend=%s:%s fd=%d",
		__func__,
		backend->dest_host,
		backend->dest_service,
		req->backend_fd);

	return 0; //success
}


int tcp_close_client_connection(int fd, dns_metric_index reason_metric)
{
	int			rc;
	fd_t			*fdte;
	dlist_t			*item;
	ioctl_opt_register_fd_t	opt;

	// close connection
	opt.fd = fd;
	opt.port = 0;
	opt.callback = NULL;
	opt.callback_arg = NULL;
	rc = pfm_vector_ioctl(priv_data.module,
		PFM_IOCTL_UNREGISTER_BACKEND_FD,
		&opt);
	if (rc && rc != ENOENT) {
		pfm_vector_log(priv_data.module,
			LOG_LEVEL_WARNING,
			"%s() unregister fd=%d failed rc=%d",
			__func__,
			fd,
			rc);
		fflush(stdout);
		assert(rc == 0);
		return rc;
	}
	(void)tcp_shutdown(fd);

	fdte = &priv_data.fd_table[fd];
	fdte->in_pb = NULL;
	fdte->in_rem = 0;
	fdte->out_pb = NULL;
	while (!dlist_is_empty(&fdte->out_qhead)) {
		item = (&fdte->out_qhead)->next;
		dlist_remove(item);
//		dlist_add(&priv_data.packetbuf_free_head, item);
		packetbuf_free((packetbuf_t*)item);
	}

	// update metrics
	metric_decr_u32(&priv_data.metric[METRIC_CONN_CURRENT]);
	metric_incr_u32(&priv_data.metric[reason_metric]);

	return 0;
}


struct timespec diff_timespec(const struct timespec *time1,
				const struct timespec *time0)
{
	assert(time1);
	assert(time0);
	struct timespec diff = {.tv_sec = time1->tv_sec - time0->tv_sec,
				.tv_nsec = time1->tv_nsec - time0->tv_nsec};
	if (diff.tv_nsec < 0) {
		diff.tv_nsec += 1000000000; // nsec/sec
		diff.tv_sec--;
	}
	return diff;
}



int udp_backend_io_callback(int fd,
	int events,
	connection_data_t *cd,
	void *arg)
{
	int			rc;
	uint16_t		plen;
	uint16_t		id;
	uint16_t		flags;
	uint16_t		q_count;
	uint16_t		a_count;
	uint16_t		ns_count;
	uint16_t		ar_count;
	int			read_len;
	char			*p;
	request_t		*req;
	packetbuf_t		*pb;
	struct timespec		ts;
	struct timespec		ts_diff;

	//************************************
	// init
	//************************************

	//************************************
	// handle socket exception
	//************************************
	if (events & EPOLLHUP) {
		pfm_vector_log(priv_data.module,
			LOG_LEVEL_ERROR,
			"HUP socket exception occurred"
			" closing UDP backend connection");
		rc = udp_shutdown(fd);
		metric_incr_u32(&priv_data.metric[METRIC_UDP_CONN_SERVER_DROP]);
		assert(0);
		return 0;
	}
	if (events & EPOLLERR) {
		pfm_vector_log(priv_data.module,
			LOG_LEVEL_ERROR,
			"%s(): ERR socket exception occurred"
			" closing UDP backend connection (may indicate a bad backend)",
			__func__);
		rc = udp_shutdown(fd);
		metric_incr_u32(&priv_data.metric[METRIC_UDP_CONN_SERVER_DROP]);
		//assert(0);
		return 0;
	}

	//************************************
	// read packet
	//************************************
	if (events & EPOLLIN) {

		//************************************
		// read packet
		//************************************
		pb = packetbuf_get_or_alloc();
		assert(pb);
		pb->sfd = fd;
		pb->saddr_len = sizeof(struct sockaddr);
		pnetbuffer_get_writable(pb->nb, &p, &plen);
		read_len = recvfrom(fd,
					p,
					(int)plen,
					0,
					&pb->saddr,
					&pb->saddr_len);
		if (read_len == -1) {
			rc = errno;
			packetbuf_free(pb);
			return rc;
		}
		rc = pnetbuffer_set_written(pb->nb, (uint16_t)read_len);
		pfm_vector_log(priv_data.module,
			LOG_LEVEL_DEBUG,
			"%s() recvfrom() fd=%d bytes=%d",
			__func__,
			fd,
			read_len);

		rc = dns_packet_read_header(pb->nb,
					&id,
					&flags,
					&q_count,
					&a_count,
					&ns_count,
					&ar_count);
		if (rc != 0) {
			pfm_vector_log(priv_data.module,
				LOG_LEVEL_WARNING,
				"%s() failed to parse packet header content"
				" fd=%d len=%d",
				__func__,
				fd,
				read_len);
			packetbuf_free(pb);
			return rc;
		}

		pfm_vector_log(priv_data.module,
			LOG_LEVEL_DEBUG,
			"%s() id=%d flags=%d q_count=%d a_count=%d",
			__func__,
			id,
			flags,
			q_count,
			a_count);

		req = request_table_get_entry(
				priv_data.fd_table[fd].ptr.request_table,
				id);
		if (!req) {
			pfm_vector_log(priv_data.module,
				LOG_LEVEL_WARNING,
				"%s() request not found in table, skip sending"
				" frontend response fd=%d id=%d",
				__func__,
				fd,
				id);
			packetbuf_free(pb);
			assert(0);
			return ENOENT;
		}
		if (req->frontend_fd <= 0) {
			pfm_vector_log(priv_data.module,
				LOG_LEVEL_NOTICE,
				"%s() req->frontend_fd not set for backend response, skip sending"
				" frontend response (backend)fd=%d  req=%p req->frontend_fd=%d id=%d",
				__func__,
				fd,
				req,
				req->frontend_fd,
				id);
//			dump_request(req);
			packetbuf_free(pb);
			//assert(req->frontend_fd != -1);
			return ENOENT;
		}
		clock_gettime(CLOCK_MONOTONIC, &ts);
		ts_diff = diff_timespec(&ts, &req->start_ts);
		pfm_vector_log(priv_data.module,
			LOG_LEVEL_INFO,
			"%s() request processing seconds=%d.%09d",
			__func__,
			ts_diff.tv_sec,
			ts_diff.tv_nsec);


		metric_incr_u32(&priv_data.metric[METRIC_ANSWER_TOTAL]);

		// TODO: add response stage answer checks
		// filter IPs?

		// forward response to source
		pnetbuffer_reset_read(pb->nb); // rewind before sending
//		if (req->frontend_fd == 0) {
//			pfm_vector_log(priv_data.module,
//				LOG_LEVEL_NOTICE,
//				"%s() dropping invalid request record"
//				" req=%p .req_id=%d .frontend_fd=%d"
//				" .backend_fd=%d",
//				__func__,
//				req,
//				req->req_id,
//				req->frontend_fd,
//				req->backend_fd);
//			dump_request(req);
//			request_reset(req);
//			return 0;
//		}
		assert(req->req_id != 0);
		rc = fd_write_or_queue_packetbuf(req->frontend_fd,
						pb,
						&req->src_sockaddr);
		if (rc != 0) {
			pfm_vector_log(priv_data.module,
				LOG_LEVEL_WARNING,
				"Failed to send response to frontend"
				" fd=%d source=%s err=%d",
				req->frontend_fd,
				inet_ntoa(((struct sockaddr_in*)&req->src_sockaddr)->sin_addr),
				rc);
		}

		pfm_vector_log(priv_data.module,
			LOG_LEVEL_DEBUG,
			"%s() wrote packet to frontend fd=%d",
			__func__,
			req->frontend_fd);
		request_table_clear_entry(
			priv_data.fd_table[fd].ptr.request_table,
			id);
	}

	//************************************
	// write packet
	//************************************
	if (events & EPOLLOUT) {
		// TODO
		// rc = fd_try_writes(fd, dest_sa);
		assert(0);
	}

	return 0;
}


filter_t *src_addr_is_filtered(struct sockaddr *sa)
{
	char			*src_ip;
	SList			*item;
	filterset_t		*fs;
	filter_t		*f;
	struct sockaddr_in	*c_sain;

	c_sain = (struct sockaddr_in *)sa;
	src_ip = inet_ntoa(c_sain->sin_addr);
	pfm_vector_log(priv_data.module,
		LOG_LEVEL_DEBUG,
		"client IP address is %s",
		src_ip);
	item = priv_data.src_filterset_list;
	while (item) {
		fs = slist_get_data(item);
		f = filterset_match(fs, src_ip);
		if (f) {
			pfm_vector_log(priv_data.module,
				LOG_LEVEL_NOTICE,
				"client IP address %s denied by filter,"
				" closing the connection",
				src_ip);
			return f;
		}
		item = slist_get_next(item);
	}
	return NULL;
}


int tcp_server_io_callback(int fd,
	int events,
	connection_data_t *cd,
	void *arg)
{
	int			rc;
	int			c_flags;
	int			c_fd;
	socklen_t		c_len;
	struct sockaddr		c_sa;
	ioctl_opt_register_fd_t	opt;
	fd_t			*fdte;
	filter_t		*f;

	// Handle new TCP connections
	metric_incr_u32(&priv_data.metric[METRIC_TCP_CONN_TOTAL]);
	c_len = sizeof(struct sockaddr);
	c_fd = accept(fd, &c_sa, &c_len);
	if (c_fd == -1) {
		rc = errno;
		pfm_vector_log(priv_data.module,
			LOG_LEVEL_WARNING,
			"Failed to accept new client connection rc=%d",
			rc);
		metric_incr_u32(&priv_data.metric[METRIC_TCP_CONN_SERVER_DROP]);
		return rc;
	}
	c_flags = fcntl(c_fd, F_GETFL, 0);
	rc = fcntl(c_fd, F_SETFL, c_flags | O_NONBLOCK);
	if (rc == -1) {
		rc = errno;
		pfm_vector_log(priv_data.module,
			LOG_LEVEL_WARNING,
			"Failed to set O_NONBLOCK on new client"
			" connection rc=%d",
			rc);
		c_flags = tcp_shutdown(c_fd);
		metric_incr_u32(&priv_data.metric[METRIC_TCP_CONN_SERVER_DROP]);
		return rc;
	}

	// source address filtering
	if (priv_data.src_filterset_list) {
		f = src_addr_is_filtered(&c_sa);
		if (f) {
			c_flags = tcp_shutdown(c_fd);
			metric_incr_u32(
			    &priv_data.metric[METRIC_TCP_CONN_SERVER_DROP]);
			return 0;
		}
	}

	opt.fd = c_fd;
	opt.port = 0;
	opt.callback = tcp_frontend_io_callback;
	opt.callback_arg = NULL;
	rc = pfm_vector_ioctl(priv_data.module,
		PFM_IOCTL_REGISTER_BACKEND_FD,
		&opt);
	if (rc) {
		pfm_vector_log(priv_data.module,
			LOG_LEVEL_ERROR,
			"Failed to register new client FD with ioloop"
			" subsystem rc=%d",
			rc);
		c_flags = tcp_shutdown(c_fd);
		metric_incr_u32(&priv_data.metric[METRIC_TCP_CONN_SERVER_DROP]);
		pfm_vector_log(priv_data.module,
			LOG_LEVEL_DEBUG,
			"%s() memory free line=%d",
			__func__,
			__LINE__);
		return rc;
	}

	fdte = &priv_data.fd_table[c_fd];
	fd_entry_init(fdte,
		DNS_CONN_TYPE_TCP,
		DNS_USAGE_TYPE_FRONTEND,
		NULL);
	fdte->saddr = c_sa;
	fdte->saddr_len = c_len;

	pfm_vector_log(priv_data.module,
		LOG_LEVEL_INFO,
		"%s() Accepted new client connection fd=%d",
		__func__,
		c_fd);
	metric_incr_u32(&priv_data.metric[METRIC_CONN_CURRENT]);

	return 0;
}


int udp_frontend_io_callback(int fd,
	int events,
	connection_data_t *cd,
	void *arg)
{
	int		rc;
	int		read_len;
	filter_t	*f;
	fd_t		*fdte;

	//************************************
	// init
	//************************************
	fdte = &priv_data.fd_table[fd];
	assert(fdte->conntype == DNS_CONN_TYPE_UDP);
	assert(fdte->usagetype == DNS_USAGE_TYPE_FRONTEND);

	//************************************
	// handle socket exception
	//************************************
	if (events & EPOLLHUP) {
		pfm_vector_log(priv_data.module,
			LOG_LEVEL_ERROR,
			"HUP socket exception occurred"
			" closing UDP frontend connection");
		rc = udp_shutdown(fd);
		pfm_vector_log(priv_data.module,
			LOG_LEVEL_DEBUG,
			"%s() memory free line=%d",
			__func__,
			__LINE__);
		metric_incr_u32(&priv_data.metric[METRIC_UDP_CONN_SERVER_DROP]);
		return 0;
	}
	if (events & EPOLLERR) {
		pfm_vector_log(priv_data.module,
			LOG_LEVEL_ERROR,
			"ERR socket exception occurred"
			" closing UDP frontend connection");
		rc = udp_shutdown(fd);
		pfm_vector_log(priv_data.module,
			LOG_LEVEL_DEBUG,
			"%s() memory free line=%d",
			__func__,
			__LINE__);
		metric_incr_u32(&priv_data.metric[METRIC_UDP_CONN_SERVER_DROP]);
		return 0;
	}

	//************************************
	// read packet
	//************************************
	if (events & EPOLLIN) {

		// start reading a new packet (UDP should always)
		fdte->in_pb = packetbuf_get_or_alloc();
		assert(fdte->in_pb);
		fdte->in_rem = 0;
		fdte->in_pb->sfd = fd;
		rc = packetbuf_try_read_from_fd(fdte->in_pb,
						fd,
						&read_len);

		if (rc && rc != EAGAIN && rc != EWOULDBLOCK) {
			// something went wrong
			pfm_vector_log(priv_data.module,
				LOG_LEVEL_WARNING,
				"Error %d occurred during read",
				rc);
			packetbuf_free(fdte->in_pb);
			fdte->in_pb = NULL;
			fdte->in_rem = 0;
//			exit(rc); // abort for now
			return rc;
		}

		// source address filtering
		if (priv_data.src_filterset_list) {
			//f = src_addr_is_filtered(&ctx.src_addr);
			f = src_addr_is_filtered(&fdte->in_pb->saddr);
			if (f) {
				// drop the request packet
				packetbuf_free(fdte->in_pb);
				fdte->in_pb = NULL;
				metric_incr_u32(
				    &priv_data.metric[METRIC_UDP_CONN_SERVER_DROP]);
				return 0;
			}
		}

		//************************************
		// process input packet
		//************************************
		pfm_vector_log(priv_data.module,
			LOG_LEVEL_INFO,
			"%s() calling vector_process_data()  fd=%d",
			__func__,
			fd);
		rc = vector_process_data(priv_data.module->manager,
				fdte->in_pb);
		if (rc) {
			// we must free this on error
			packetbuf_free(fdte->in_pb);
		}
		fdte->in_pb = NULL; // the pb is freed once sent
	}

	//************************************
	// write packet
	//************************************
	if (events & EPOLLOUT) {
		// TODO
		// rc = fd_try_writes(fd, dest_sa);
		assert(0);
	}

	return 0;
}


int tcp_frontend_io_callback(int fd,
	int events,
	connection_data_t *cd,
	void *arg)
{
	int		rc;
	fd_t		*fdte;
	packetbuf_t	*pb = NULL;

	//************************************
	// init
	//************************************
	fdte = &priv_data.fd_table[fd];
	assert(fdte->conntype == DNS_CONN_TYPE_TCP);
	assert(fdte->usagetype == DNS_USAGE_TYPE_FRONTEND);

	//************************************
	// handle socket exception
	//************************************
	if (events & EPOLLHUP) {
		pfm_vector_log(priv_data.module,
			LOG_LEVEL_WARNING,
			"HUP socket exception occurred"
			" closing client connection");
		rc = tcp_close_client_connection(fd,
				METRIC_TCP_CONN_SERVER_DROP);
		return 0;
	}
	if (events & EPOLLERR) {
		pfm_vector_log(priv_data.module,
			LOG_LEVEL_WARNING,
			"ERR socket exception occurred"
			" closing client connection");
		rc = tcp_close_client_connection(fd,
				METRIC_TCP_CONN_SERVER_DROP);
		return 0;
	}

	//************************************
	// read packet
	//************************************
	if (events & EPOLLIN) {
		
//		if (!fdte->in_pb) {
//			pb = packetbuf_get_or_alloc();
//			assert(pb != NULL);
//			fdte->in_pb = pb;
//		} else {
//			pb = fdte->in_pb;
//		}
		rc = tcp_try_read_packet(fd, &pb);
//		pfm_vector_log(priv_data.module,
//			LOG_LEVEL_ERROR,
//			"%s() fd=%d tcp_try_read_packet() rc=%d",
//			__func__,
//			fd,
//			rc);
		if (rc == EAGAIN) {
			goto pollin_end;
			return 0;
		}
		if (rc) {
			// something went wrong
			pfm_vector_log(priv_data.module,
				LOG_LEVEL_INFO,
				"Closing client TCP connection");
			(void)tcp_close_client_connection(fd,
					METRIC_TCP_CONN_SERVER_DROP);
			goto pollin_end;
			return rc;
		}

		// TCP: use source info from accept()
		pb->saddr = fdte->saddr;
		pb->saddr_len = fdte->saddr_len;
		
		// pb = fdte->in_pb = a complete request
		// and the read position is at the start of
		// the DNS request (past TCP length header)

		//************************************
		// process input packet
		//************************************
		pfm_vector_log(priv_data.module,
			LOG_LEVEL_INFO,
			"%s() calling vector_process_data()  fd=%d",
			__func__,
			fd);
		rc = vector_process_data(priv_data.module->manager,
				pb);
		if (rc) {
			// we must free this on error
			packetbuf_free(fdte->in_pb);
		}
		fdte->in_pb = NULL; // the pb is freed once sent
	}
pollin_end:

	//************************************
	// write packet
	//************************************
	if (events & EPOLLOUT) {
		// TODO
		pfm_vector_log(priv_data.module,
			LOG_LEVEL_WARNING,
			"%s() EPOLLOUT received, calling fd_try_writes() fd=%d",
			__func__,
			fd);
		rc = fd_try_writes(fd, NULL);
		if (rc)
			assert(rc == 0);
		return 0;
	}

	return 0;
}


int tcp_backend_io_callback(int fd,
	int events,
	connection_data_t *cd,
	void *arg)
{
	int			rc;
	uint16_t		id;
	uint16_t		flags;
	uint16_t		q_count;
	uint16_t		a_count;
	uint16_t		ns_count;
	uint16_t		ar_count;
	request_t		*req;
	fd_t			*fdte;
	backend_t		*backend;
	struct timespec		ts;
	struct timespec		ts_diff;
	packetbuf_t		*pb = NULL;

	//************************************
	// init
	//************************************
	fdte = &priv_data.fd_table[fd];
	pfm_vector_log(priv_data.module,
		LOG_LEVEL_DEBUG,
		"%s() io event for fd=%d events=%d fdte=%p",
		__func__,
		fd,
		events,
		fdte);

	//************************************
	// handle socket exception
	//************************************
	if (events & EPOLLHUP) {
		pfm_vector_log(priv_data.module,
			LOG_LEVEL_WARNING,
			"HUP socket exception occurred"
			" replacing TCP backend connection");
		// replace this connection
		backend = (backend_t*)fdte->backend;
		rc = backend_replace_connection(backend, fd);
		metric_incr_u32(
			&priv_data.metric[METRIC_TCP_BACKEND_RECONNECT]);
		// TODO move queued packets to another backend
		return 0;
	}
	if (events & EPOLLERR) {
		pfm_vector_log(priv_data.module,
			LOG_LEVEL_WARNING,
			"ERR socket exception occurred"
			" replacing TCP backend connection"
			" (may indicate bad backend)");
		// replace this connection
		backend = (backend_t*)fdte->backend;
		rc = backend_replace_connection(backend, fd);
		metric_incr_u32(
			&priv_data.metric[METRIC_TCP_BACKEND_RECONNECT]);
		// TODO move queued packets to another backend
		return 0;
	}

	//************************************
	// read packet
	//************************************
	if (events & EPOLLIN) {
//		if (!fdte->in_pb) {
//			pb = packetbuf_get_or_alloc();
//			assert(pb != NULL);
//			fdte->in_pb = pb;
//		} else {
//			pb = fdte->in_pb;
//		}

		rc = tcp_try_read_packet(fd, &pb);
		if (rc == EAGAIN)
			return 0;
		if (rc) {
			// something went wrong
			pfm_vector_log(priv_data.module,
				LOG_LEVEL_DEBUG,
				"%s() Socket error occurred rc=%d"
				" replacing TCP backend connection",
				__func__,
				rc);
			// replace this connection and
			backend = (backend_t*)fdte->backend;
			rc = backend_replace_connection(backend, fd);
			metric_incr_u32(
				&priv_data.metric[METRIC_TCP_BACKEND_RECONNECT]);
			// TODO move queued packets to another backend
			goto pollin_end;
			return rc;
		}
		// pb = fdte->in_pb = a complete request
		// and the read position is at the start of
		// the DNS request (past TCP length header)

		rc = dns_packet_read_header(pb->nb,
					&id,
					&flags,
					&q_count,
					&a_count,
					&ns_count,
					&ar_count);
		if (rc != 0) {
			pfm_vector_log(priv_data.module,
				LOG_LEVEL_ERROR,
				"%s() Failed to read DNS header",
				__func__,
				fd);
			// TODO handle this error
			packetbuf_free(pb);
//			fdte->in_pb = NULL;
			goto pollin_end;
			return rc;
		}

		pfm_vector_log(priv_data.module,
			LOG_LEVEL_DEBUG,
			"%s() id=%d flags=%d q_count=%d a_count=%d",
			__func__,
			id,
			flags,
			q_count,
			a_count);
		
		// retrieve request object
		req = request_table_get_entry(
				fdte->ptr.request_table,
				id);
		if (!req) {
			pfm_vector_log(priv_data.module,
				LOG_LEVEL_WARNING,
				"%s() request not found in table, skip sending"
				" frontend response fd=%d id=%d",
				__func__,
				fd,
				id);
			packetbuf_free(pb);
//			fdte->in_pb = NULL;
			goto pollin_end;
			return ENOENT;
		}
		clock_gettime(CLOCK_MONOTONIC, &ts);
		ts_diff = diff_timespec(&ts, &req->start_ts);
		pfm_vector_log(priv_data.module,
			LOG_LEVEL_INFO,
			"%s() request processing seconds=%d.%09d",
			__func__,
			ts_diff.tv_sec,
			ts_diff.tv_nsec);

		if (req->frontend_fd <= 0) {
			pfm_vector_log(priv_data.module,
				LOG_LEVEL_NOTICE,
				"%s() dropping invalid response record"
				" req=%p .req_id=%d .frontend_fd=%d"
				" .backend_fd=%d",
				__func__,
				req,
				req->req_id,
				req->frontend_fd,
				req->backend_fd);
//			dump_request(req);
			packetbuf_free(pb);
			fdte->in_pb = NULL;
			goto pollin_end;
		}
		assert(req->frontend_fd > 0);
		pb->req = req;

		metric_incr_u32(&priv_data.metric[METRIC_ANSWER_TOTAL]);

		// TODO: add response stage answer checks
		// filter IPs?

		// forward response to frontend
		pnetbuffer_reset_read(pb->nb); // rewind before sending
		rc = fd_write_or_queue_packetbuf(req->frontend_fd,
						pb,
						&req->src_sockaddr);
		if (rc != 0) {
			pfm_vector_log(priv_data.module,
				LOG_LEVEL_WARNING,
				"Failed to send frontend response"
				"fd=%d source=%s err=%d",
				req->frontend_fd,
				inet_ntoa(((struct sockaddr_in*)&req->src_sockaddr)->sin_addr),
				rc);
			packetbuf_free(pb);
//			fdte->in_pb = NULL;
			request_reset(req);
			goto pollin_end;
			return rc;
		}

		pfm_vector_log(priv_data.module,
			LOG_LEVEL_DEBUG,
			"%s() wrote packet to frontend fd=%d id=%d",
			__func__,
			rc,
			req->frontend_fd,
			id);
		request_table_clear_entry(
			fdte->ptr.request_table,
			id);
		fdte->in_pb = NULL; // the pb is freed once sent
	}
pollin_end:

	//************************************
	// write packet
	//************************************
	if (events & EPOLLOUT) {
		// TODO
		pfm_vector_log(priv_data.module,
			LOG_LEVEL_WARNING,
			"%s() EPOLLOUT received, calling fd_try_writes() fd=%d",
			__func__,
			fd);
		rc = fd_try_writes(fd, NULL);
		if (rc)
			assert(rc == 0);
	}

	return 0;
}


extern int backend_io_callback(int fd,
	int events,
	connection_data_t *cd,
	void *arg)
{
	int		rc;
	fd_t		*fdte;

	fdte = &priv_data.fd_table[fd];
	pfm_vector_log(priv_data.module,
		LOG_LEVEL_DEBUG,
		"%s() io event for fd=%d events=%d fdte=%p",
		__func__,
		fd,
		events,
		fdte);
	pfm_vector_log(priv_data.module,
		LOG_LEVEL_DEBUG,
		"%s() fdte->conntype=%d fdte->usagetype=%d fdte->ptr->ptr=%p",
		__func__,
		fdte->conntype,
		fdte->usagetype,
		fdte->ptr.ptr);
	dump_fdte(fdte);
	if (priv_data.fd_table[fd].conntype == DNS_CONN_TYPE_UDP) {
		rc = udp_backend_io_callback(fd, events, cd, arg);
	} else if (priv_data.fd_table[fd].conntype == DNS_CONN_TYPE_TCP) {
		rc = tcp_backend_io_callback(fd, events, cd, arg);
	} else {
		pfm_vector_log(priv_data.module,
			LOG_LEVEL_DEBUG,
			"%s() unhandled conntype fd=%d conntype=%d EXITING",
			__func__,
			fd,
			fdte->conntype);
		exit(EIO);
		return 0;
	}
	return rc;
}

