/************************************************************************/
/* pfm_dns.h - DNS packet filter module					*/
/************************************************************************/
/*
 *  (c)2021,2022 Todd Shadburn <tshadburn@opencomputingsolutions.com>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
/************************************************************************/


#ifndef _PFM_DNS_H_
#define _PFM_DNS_H_


#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <string.h>
#include <errno.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <cutil/cutil.h>
#include <pcre.h>

#include <pnetbuffer.h>
#include <logging.h>
#include <udp.h>
#include <tcp.h>
#include <pfm.h>
#include <metric.h>
#include <helpers.h>
#include <backend.h>
#include <requests.h>



#define PACKET_BUFFER_SIZE	16384
#define PFM_DNS_PORT_DEFAULT	53
#define RESPONSE_ARRAY_SIZE	4
#define RESPONSE_STRING_SIZE	256


#define VOTE_PASS			0
#define VOTE_DENY			1
#define VOTE_DROP			2
#define VOTE_LOG			3


#define DNS_F_QR			0x8000
#define DNS_F_AA			0x0400
#define DNS_F_TC			0x0200
#define DNS_F_RD			0x0100
#define DNS_F_RA			0x0080

#define DNS_F_OP_0			0x0080
#define DNS_F_OP_1			0x0100
#define DNS_F_OP_2			0x0200
#define DNS_F_OP_3			0x0400

#define DNS_F_Z_0			0x0010
#define DNS_F_Z_1			0x0020
#define DNS_F_Z_2			0x0040

#define DNS_F_RC_0			0x0001
#define DNS_F_RC_1			0x0002
#define DNS_F_RC_2			0x0004
#define DNS_F_RC_3			0x0008
#define DNS_F_RC_SUCCESS		0
#define DNS_F_RC_FMTERROR		1
#define DNS_F_RC_SRVERROR		2
#define DNS_F_RC_NXDOMAIN		3
#define DNS_F_RC_NOIMPL			4
#define DNS_F_RC_REFUSED		5


#define HASH_TABLE_SIZE_DEFAULT		2048


typedef enum {
	METRIC_QUERY_TOTAL	= 0,
	METRIC_QUERY_FILTERED,
	METRIC_ANSWER_TOTAL,
	METRIC_ANSWER_FILTERED,
	METRIC_TCP_CONN_TOTAL,
	METRIC_TCP_CONN_SERVER_DROP,
	METRIC_TCP_BACKEND_RECONNECT,
	METRIC_UDP_CONN_SERVER_DROP,
	METRIC_CONN_CURRENT,
	_METRIC_COUNT,
} dns_metric_index;


typedef enum {
	DNS_USAGE_TYPE_SERVER	= 0,
	DNS_USAGE_TYPE_FRONTEND,
	DNS_USAGE_TYPE_BACKEND,
	_DNS_USAGE_TYPE_MAX,
} dns_usage_type;

typedef enum {
	DNS_CONN_TYPE_UDP	= 0,
	DNS_CONN_TYPE_TCP,
	_DNS_CONN_TYPE_MAX,
} dns_conn_type;


typedef enum {
	RC_BACKEND_SUCCESS		= 0,
	RC_BACKEND_NO_PROTO,
	RC_BACKEND_NO_CONNS,
} backend_conn_err_t;


typedef struct _packetbuf_t packetbuf_t;
struct _packetbuf_t {
	dlist_t		list;

	pnetbuffer_t	*nb;
	char		nb_buffer[PACKET_BUFFER_SIZE];

	int		sfd;
	struct sockaddr	saddr;
	socklen_t	saddr_len;

	request_t	*req;
};


typedef struct _fd_t fd_t;
struct _fd_t {
	dns_conn_type	conntype;
	dns_usage_type	usagetype;

	// backend pointer (only for backend FDs)
	void		*backend;

	// FD state pointer (request_table_t, etc)
	union {
		void			*ptr;
		request_table_t		*request_table;
	} ptr;

	packetbuf_t	*in_pb;	// curr. inbound packet
	uint32_t	in_rem;	// remaining bytes to read

	packetbuf_t	*out_pb;  // curr. outbound packet
	dlist_t		out_qhead; // queued outbound packets

	// TCP frontend connection source info
	struct sockaddr	saddr;
	socklen_t	saddr_len;
};



//typedef struct _dnsconn_t dnsconn_t;
//struct _dnsconn_t {
//	dns_conn_type	conntype;
//
//	int		frontend_fd;
//	struct sockaddr	src_sa;
//	socklen_t	src_salen;
//};



#define DNS_BACKENDS_MAX		16
#define DNS_FD_TABLE_SIZE_DEFAULT	1024
typedef struct _pfm_dns_t pfm_dns_t;
struct _pfm_dns_t {
	pfm_module_t	*module;	// passed in module_open()

	uint64_t	policy_bits;

	char		**command_array;
	char		**response_array;

	metric_t	metric[_METRIC_COUNT];

	SList		*filterset_list;

	SList		*q_filterset_list;
	SList		*an_filterset_list;
	SList		*src_filterset_list;

	char		*service_bind_address;

	// backend count by protocol
	int		backend_count[_DNS_CONN_TYPE_MAX];
	// backend array by protocol
	backend_t	*backend[_DNS_CONN_TYPE_MAX][DNS_BACKENDS_MAX];

	int		fd_table_size;
	fd_t		*fd_table;

	dlist_t		packetbuf_free_head;
};


// policy bits management
typedef enum {
	POLICY_NOOP				= 0,
	POLICY_DENY_MULTIPLE_QUERIES,

	_POLICY_MAXIMUM,
} policy_bit_t;

typedef struct _policy_item_t policy_item_t;
struct _policy_item_t {
	char		*name;
	policy_bit_t	bit; // policy_bits |= (1<<bit)
};

//policy_item_t policy_table[] = {
//	{"deny-multiquery",	POLICY_DENY_MULTIPLE_QUERIES},
//	{NULL,			POLICY_NOOP},
//};



void fd_entry_init(fd_t *fde,
		dns_conn_type conntype,
		dns_usage_type usagetype,
		void *dataptr);
void fd_entry_clear(fd_t *fde);


int backend_io_callback(int fd,
	int events,
	connection_data_t *cd,
	void *arg);


#endif



